from __future__ import annotations

from functools import lru_cache, partial
from typing import TYPE_CHECKING, List, Optional, Set, Tuple, Union

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import AstNodeId, SolcIdentifier
from woke.ir.declarations.abc import DeclarationAbc
from woke.ir.declarations.variable_declaration import VariableDeclaration
from woke.ir.enums import GlobalSymbolsEnum, ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.reference_resolver import CallbackParams
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from woke.ir.meta.import_directive import ImportDirective
    from woke.ir.meta.source_unit import SourceUnit


class Identifier(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcIdentifier
    _parent: SolidityAbc  # TODO: make this more specific

    _name: str
    _overloaded_declarations: List[AstNodeId]
    _referenced_declaration_id: Optional[AstNodeId]

    def __init__(
        self, init: IrInitTuple, identifier: SolcIdentifier, parent: SolidityAbc
    ):
        from ..meta.import_directive import ImportDirective

        super().__init__(init, identifier, parent)
        self._name = identifier.name
        self._overloaded_declarations = list(identifier.overloaded_declarations)
        self._referenced_declaration_id = identifier.referenced_declaration
        if self._referenced_declaration_id is None:
            assert isinstance(self._parent, ImportDirective)
        init.reference_resolver.register_post_process_callback(
            self._post_process, priority=-1
        )

    def _post_process(self, callback_params: CallbackParams):
        from ..meta.import_directive import ImportDirective

        assert self._referenced_declaration_id is not None
        if self._referenced_declaration_id < 0:
            global_symbol = GlobalSymbolsEnum(self._referenced_declaration_id)
            self._reference_resolver.register_global_symbol_reference(
                global_symbol, self
            )
            self._reference_resolver.register_destroy_callback(
                self.file, partial(self._destroy, global_symbol)
            )
        else:
            node = self._reference_resolver.resolve_node(
                self._referenced_declaration_id, self._cu_hash
            )

            if isinstance(node, DeclarationAbc):
                node.register_reference(self)
                self._reference_resolver.register_destroy_callback(
                    self.file, partial(self._destroy, node)
                )
            elif isinstance(node, ImportDirective):
                # make this node to reference the source unit directly
                assert node.unit_alias is not None
                source_unit = callback_params.source_units[node.imported_file]
                node_path_order = self._reference_resolver.get_node_path_order(
                    AstNodeId(source_unit.ast_node_id),
                    source_unit.cu_hash,
                )
                self._referenced_declaration_id = (
                    self._reference_resolver.get_ast_id_from_cu_node_path_order(
                        node_path_order, self.cu_hash
                    )
                )
            else:
                raise TypeError(f"Unexpected type: {type(node)}")

    def _destroy(
        self, referenced_declaration: Union[GlobalSymbolsEnum, DeclarationAbc]
    ) -> None:
        if isinstance(referenced_declaration, GlobalSymbolsEnum):
            self._reference_resolver.unregister_global_symbol_reference(
                referenced_declaration, self
            )
        elif isinstance(referenced_declaration, DeclarationAbc):
            referenced_declaration.unregister_reference(self)
        else:
            raise TypeError(f"Unexpected type: {type(referenced_declaration)}")

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def name(self) -> str:
        return self._name

    @property
    def overloaded_declarations(self) -> Tuple[DeclarationAbc, ...]:
        overloaded_declarations = []
        for overloaded_declaration_id in self._overloaded_declarations:
            if overloaded_declaration_id < 0:
                continue

            overloaded_declaration = self._reference_resolver.resolve_node(
                overloaded_declaration_id, self._cu_hash
            )
            assert isinstance(overloaded_declaration, DeclarationAbc)
            overloaded_declarations.append(overloaded_declaration)
        return tuple(overloaded_declarations)

    @property
    def referenced_declaration(
        self,
    ) -> Union[DeclarationAbc, GlobalSymbolsEnum, SourceUnit]:
        from ..meta.source_unit import SourceUnit

        assert self._referenced_declaration_id is not None
        if self._referenced_declaration_id < 0:
            return GlobalSymbolsEnum(self._referenced_declaration_id)

        node = self._reference_resolver.resolve_node(
            self._referenced_declaration_id, self._cu_hash
        )
        assert isinstance(
            node, (DeclarationAbc, SourceUnit)
        ), f"Unexpected type: {type(node)}\n{node.source}\n{self.source}\n{self.file}"
        return node

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        referenced_declaration = self.referenced_declaration
        return (
            isinstance(referenced_declaration, VariableDeclaration)
            and referenced_declaration.is_state_variable
        )

    @property
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return set()



from enum import IntEnum, IntFlag

from woke.utils import StrEnum


class GlobalSymbolsEnum(IntEnum):
    """
    Global symbols of the Solidity language. Symbols with identifiers from `-1` to `-99` are codified by the compiler and can only be referenced by [Identifier][woke.ir.expressions.identifier.Identifier] nodes.
    Other symbols are not officially codified by the compiler, but Woke also defines identifiers for them. These symbols can only be referenced by [MemberAccess][woke.ir.expressions.member_access.MemberAccess] nodes.
    See the [Solidity docs](https://docs.soliditylang.org/en/latest/units-and-global-variables.html#special-variables-and-functions) for (an incomplete) list of global symbols and their descriptions.
    """

    ABI = -1
    ADDMOD = -2
    ASSERT = -3
    BLOCK = -4
    BLOCKHASH = -5
    ECRECOVER = -6
    GASLEFT = -7
    KECCAK256 = -8
    MSG = -15
    MULMOD = -16
    NOW = -17
    REQUIRE = -18
    REVERT = -19
    RIPEMD160 = -20
    SELFDESTRUCT = -21
    SHA256 = -22
    SHA3 = -23
    SUICIDE = -24
    SUPER = -25
    TX = -26
    TYPE = -27
    THIS = -28

    BLOCK_BASEFEE = -100
    BLOCK_CHAINID = -101
    BLOCK_COINBASE = -102
    BLOCK_DIFFICULTY = -103
    BLOCK_GASLIMIT = -104
    BLOCK_NUMBER = -105
    BLOCK_TIMESTAMP = -106
    BLOCK_PREVRANDAO = -107

    MSG_DATA = -200
    MSG_SENDER = -201
    MSG_SIG = -202
    MSG_VALUE = -203

    TX_GASPRICE = -300
    TX_ORIGIN = -301

    ABI_DECODE = -400
    ABI_ENCODE = -401
    ABI_ENCODE_PACKED = -402
    ABI_ENCODE_WITH_SELECTOR = -403
    ABI_ENCODE_WITH_SIGNATURE = -404
    ABI_ENCODE_CALL = -405

    BYTES_CONCAT = -500
    BYTES_LENGTH = -501
    BYTES_PUSH = -502

    STRING_CONCAT = -600

    ADDRESS_BALANCE = -700
    ADDRESS_CODE = -701
    ADDRESS_CODEHASH = -702
    ADDRESS_TRANSFER = -703
    ADDRESS_SEND = -704
    ADDRESS_CALL = -705
    ADDRESS_DELEGATECALL = -706
    ADDRESS_STATICCALL = -707

    # available for contracts and interfaces
    TYPE_NAME = -800
    TYPE_CREATION_CODE = -801
    TYPE_RUNTIME_CODE = -802
    # available for interfaces only
    TYPE_INTERFACE_ID = -803
    # available for integers
    TYPE_MIN = -804
    TYPE_MAX = -805

    ARRAY_LENGTH = -900
    ARRAY_PUSH = -901
    ARRAY_POP = -902

    FUNCTION_SELECTOR = -1000
    FUNCTION_VALUE = -1001
    FUNCTION_GAS = -1002
    FUNCTION_ADDRESS = -1003

    USER_DEFINED_VALUE_TYPE_WRAP = -1100
    USER_DEFINED_VALUE_TYPE_UNWRAP = -1101


# Contracts
class ContractKind(StrEnum):
    """
    Kind of a [ContractDefinition][woke.ir.declarations.contract_definition.ContractDefinition] declaration node.
    """

    CONTRACT = "contract"
    INTERFACE = "interface"
    LIBRARY = "library"


# State variables
class Mutability(StrEnum):
    """
    Mutability of a [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration] declaration node.
    """

    MUTABLE = "mutable"
    IMMUTABLE = "immutable"
    CONSTANT = "constant"


# Functions
class FunctionKind(StrEnum):
    """
    Kind of a [FunctionDefinition][woke.ir.declarations.function_definition.FunctionDefinition] declaration node.
    """

    FUNCTION = "function"
    RECEIVE = "receive"
    CONSTRUCTOR = "constructor"
    FALLBACK = "fallback"
    FREE_FUNCTION = "freeFunction"
    """
    Function defined outside of a contract.
    """


class Visibility(StrEnum):
    """
    Visibility of:

    - [FunctionTypeName][woke.ir.type_names.function_type_name.FunctionTypeName] type name,
    - [FunctionDefinition][woke.ir.declarations.function_definition.FunctionDefinition], [ModifierDefinition][woke.ir.declarations.modifier_definition.ModifierDefinition], [StructDefinition][woke.ir.declarations.struct_definition.StructDefinition] and [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration] declarations.
    """

    EXTERNAL = "external"
    PUBLIC = "public"
    INTERNAL = "internal"
    PRIVATE = "private"


class StateMutability(StrEnum):
    """
    State mutability of:

    - [Function][woke.ir.types.Function] type,
    - [FunctionDefinition][woke.ir.declarations.function_definition.FunctionDefinition] declaration,
    - [ElementaryTypeName][woke.ir.type_names.elementary_type_name.ElementaryTypeName] and [FunctionTypeName][woke.ir.type_names.function_type_name.FunctionTypeName] type names.

    In the case of [ElementaryTypeName][woke.ir.type_names.elementary_type_name.ElementaryTypeName], the state mutability is specified only for the `address` type and can be either [NONPAYABLE][woke.ir.enums.StateMutability.NONPAYABLE] or [PAYABLE][woke.ir.enums.StateMutability.PAYABLE].
    """

    PAYABLE = "payable"
    PURE = "pure"
    NONPAYABLE = "nonpayable"
    VIEW = "view"


class ModifierInvocationKind(StrEnum):
    """
    Kind of a [ModifierInvocation][woke.ir.meta.modifier_invocation.ModifierInvocation] meta node.
    """

    MODIFIER_INVOCATION = "modifierInvocation"
    BASE_CONSTRUCTOR_SPECIFIER = "baseConstructorSpecifier"


# Literals
class LiteralKind(StrEnum):
    """
    Kind of a [Literal][woke.ir.expressions.literal.Literal] expression node.
    """

    BOOL = "bool"
    NUMBER = "number"
    STRING = "string"
    HEX_STRING = "hexString"
    UNICODE_STRING = "unicodeString"


class YulLiteralValueKind(StrEnum):
    """
    Kind of a [YulLiteral][woke.ir.yul.literal.YulLiteral] node.
    """

    NUMBER = "number"
    STRING = "string"
    BOOL = "bool"


# Function call
class FunctionCallKind(StrEnum):
    """
    Kind of a [FunctionCall][woke.ir.expressions.function_call.FunctionCall] expression node.
    """

    FUNCTION_CALL = "functionCall"
    """
    Represents also an error call, event call and [NewExpression][woke.ir.expressions.new_expression.NewExpression] call.
    """
    TYPE_CONVERSION = "typeConversion"
    STRUCT_CONSTRUCTOR_CALL = "structConstructorCall"


# Compound types
class DataLocation(StrEnum):
    """
    Data location of a [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration] node.
    It also specifies the data location of the following types:

    - [Array][woke.ir.types.Array],
    - [Bytes][woke.ir.types.Bytes],
    - [String][woke.ir.types.String],
    - [Struct][woke.ir.types.Struct].
    """

    CALLDATA = "calldata"
    DEFAULT = "default"
    """
    Set only in [VariableDeclaration][woke.ir.declarations.variable_declaration.VariableDeclaration] nodes when the data location is not specified (and the compiler even does not allow it).
    """
    MEMORY = "memory"
    STORAGE = "storage"


# Operations
class UnaryOpOperator(StrEnum):
    """
    Unary operation operator used in an [UnaryOperation][woke.ir.expressions.unary_operation.UnaryOperation] expression.
    """

    PLUS_PLUS = r"++"
    MINUS_MINUS = r"--"
    MINUS = r"-"
    NOT = r"!"
    TILDE = r"~"
    DELETE = "delete"


class AssignmentOperator(StrEnum):
    """
    Assignment operator used in an [Assignment][woke.ir.expressions.assignment.Assignment] expression.
    """

    EQUALS = r"="
    PLUS_EQUALS = r"+="
    MINUS_EQUALS = r"-="
    TIMES_EQUALS = r"*="
    DIVIDE_EQUALS = r"/="
    MODULO_EQUALS = r"%="
    OR_EQUALS = r"|="
    AND_EQUALS = r"&="
    XOR_EQUALS = r"^="
    RIGHT_SHIFT_EQUALS = r">>="
    LEFT_SHIFT_EQUALS = r"<<="


class BinaryOpOperator(StrEnum):
    """
    Binary operation operator used in a [BinaryOperation][woke.ir.expressions.binary_operation.BinaryOperation] expression.
    """

    PLUS = r"+"
    MINUS = r"-"
    TIMES = r"*"
    DIVIDE = r"/"
    MODULO = r"%"
    EXP = r"**"
    BOOLEAN_AND = r"&&"
    BOOLEAN_OR = r"||"
    NEQ = r"!="
    EQ = r"=="
    LT = r"<"
    LTE = r"<="
    GT = r">"
    GTE = r">="
    XOR = r"^"
    BITWISE_AND = r"&"
    BITWISE_OR = r"|"
    LEFT_SHIFT = r"<<"
    RIGHT_SHIFT = r">>"


# Inline assembly
class InlineAssemblyEvmVersion(StrEnum):
    HOMESTEAD = "homestead"
    TANGERINE_WHISTLE = "tangerineWhistle"
    SPURIOUS_DRAGON = "spuriousDragon"
    BYZANTIUM = "byzantium"
    CONSTANTINOPLE = "constantinople"
    PETERSBURG = "petersburg"
    ISTANBUL = "istanbul"
    BERLIN = "berlin"
    LONDON = "london"
    PARIS = "paris"
    SHANGHAI = "shanghai"


class InlineAssemblySuffix(StrEnum):
    SLOT = "slot"
    OFFSET = "offset"
    LENGTH = "length"
    ADDRESS = "address"
    SELECTOR = "selector"


class InlineAssemblyFlag(StrEnum):
    MEMORY_SAFE = "memory-safe"


class ModifiesStateFlag(IntFlag):
    """
    Flag enum describing how an expression ([ExpressionAbc][woke.ir.expressions.abc.ExpressionAbc]) or statement ([StatementAbc][woke.ir.statements.abc.StatementAbc]) modifies the blockchain state.
    """

    MODIFIES_STATE_VAR = 1
    EMITS = 2
    SENDS_ETHER = 4
    DEPLOYS_CONTRACT = 8
    SELFDESTRUCTS = 16
    PERFORMS_CALL = 32
    PERFORMS_DELEGATECALL = 64
    CALLS_UNIMPLEMENTED_NONPAYABLE_FUNCTION = 128
    CALLS_UNIMPLEMENTED_PAYABLE_FUNCTION = 256

    def __repr__(self):
        if self.value == 0:
            return f"{self.__class__.__name__}(0)"
        flags = [f for f in self.__class__ if f in self]
        return " | ".join(f.name or "" for f in flags)

    __str__ = __repr__


class FunctionTypeKind(StrEnum):
    """
    Kind of a [Function][woke.ir.types.Function] type.
    """

    DECLARATION = "declaration"
    INTERNAL = "internal"
    EXTERNAL = "external"
    DELEGATE_CALL = "delegatecall"
    BARE_CALL = "barecall"
    BARE_CALL_CODE = "barecallcode"
    BARE_DELEGATE_CALL = "baredelegatecall"
    BARE_STATIC_CALL = "barestaticcall"
    CREATION = "creation"
    SEND = "send"
    TRANSFER = "transfer"
    KECCAK256 = "keccak256"
    SELFDESTRUCT = "selfdestruct"
    REVERT = "revert"
    EC_RECOVER = "ecrecover"
    SHA256 = "sha256"
    RIPEMD160 = "ripemd160"
    LOG0 = "log0"
    LOG1 = "log1"
    LOG2 = "log2"
    LOG3 = "log3"
    LOG4 = "log4"
    GAS_LEFT = "gasleft"
    EVENT = "event"
    ERROR = "error"
    WRAP = "wrap"
    UNWRAP = "unwrap"
    SET_GAS = "setgas"
    SET_VALUE = "setvalue"
    BLOCK_HASH = "blockhash"
    ADD_MOD = "addmod"
    MUL_MOD = "mulmod"
    ARRAY_PUSH = "arraypush"
    ARRAY_POP = "arraypop"
    BYTE_ARRAY_PUSH = "bytearraypush"
    BYTES_CONCAT = "bytesconcat"
    STRING_CONCAT = "stringconcat"
    OBJECT_CREATION = "objectcreation"
    ASSERT = "assert"
    REQUIRE = "require"
    ABI_ENCODE = "abiencode"
    ABI_ENCODE_PACKED = "abiencodepacked"
    ABI_ENCODE_WITH_SELECTOR = "abiencodewithselector"
    ABI_ENCODE_CALL = "abiencodecall"
    ABI_ENCODE_WITH_SIGNATURE = "abiencodewithsignature"
    ABI_DECODE = "abidecode"
    META_TYPE = "metatype"


class MagicTypeKind(StrEnum):
    """
    Kind of a [Magic][woke.ir.types.Magic] type.
    """

    BLOCK = "block"
    MESSAGE = "message"
    TRANSACTION = "transaction"
    ABI = "abi"
    META_TYPE = "meta_type"



from __future__ import annotations

import re
import typing as typ
from abc import ABC, abstractmethod

if typ.TYPE_CHECKING:
    from woke.ir.declarations.contract_definition import ContractDefinition
    from woke.ir.declarations.enum_definition import EnumDefinition
    from woke.ir.declarations.struct_definition import StructDefinition
    from woke.ir.declarations.user_defined_value_type_definition import (
        UserDefinedValueTypeDefinition,
    )

from woke.ir.ast import AstNodeId
from woke.ir.enums import DataLocation, FunctionTypeKind, MagicTypeKind, StateMutability
from woke.ir.reference_resolver import ReferenceResolver
from woke.utils.string import StringReader

NUMBER_RE = re.compile(r"^(?P<number>[0-9]+)")
HEX_RE = re.compile(r"^(?P<hex>[0-9a-fA-F]+)")
IDENTIFIER_RE = re.compile(r"^\$_(?P<identifier>[a-zA-Z\$_][a-zA-Z0-9\$_]*)_\$")


class TypeAbc(ABC):
    """
    Abstract base class for all types.
    """

    @classmethod
    def from_type_identifier(
        cls,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ) -> typ.Optional["TypeAbc"]:
        if type_identifier.startswith("t_address"):
            return Address(type_identifier)
        elif type_identifier.startswith("t_bool"):
            return Bool(type_identifier)
        elif type_identifier.startswith("t_int"):
            return Int(type_identifier)
        elif type_identifier.startswith("t_uint"):
            return UInt(type_identifier)
        elif type_identifier.startswith("t_stringliteral_"):
            return StringLiteral(type_identifier)
        elif type_identifier.startswith("t_string_"):
            return String(type_identifier)
        elif type_identifier.startswith("t_bytes_"):
            return Bytes(type_identifier)
        # must go after t_bytes_ !!
        elif type_identifier.startswith("t_bytes"):
            return FixedBytes(type_identifier)
        elif type_identifier.startswith("t_function"):
            return Function(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_tuple"):
            return Tuple(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_type"):
            return Type(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_rational"):
            return Rational(type_identifier)
        elif type_identifier.startswith("t_modifier"):
            return Modifier(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_array"):
            return Array(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_mapping"):
            return Mapping(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_contract") or type_identifier.startswith(
            "t_super"
        ):
            return Contract(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_struct"):
            return Struct(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_enum"):
            return Enum(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_magic"):
            return Magic(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_userDefinedValueType"):
            return UserDefinedValueType(type_identifier, reference_resolver, cu_hash)
        elif type_identifier.startswith("t_module"):
            return Module(type_identifier)
        elif type_identifier.startswith("t_fixed"):
            return Fixed(type_identifier)
        elif type_identifier.startswith("t_ufixed"):
            return UFixed(type_identifier)
        else:
            return None

    @abstractmethod
    def abi_type(self) -> str:
        ...


def _parse_list(
    type_identifier: StringReader, reference_resolver: ReferenceResolver, cu_hash: bytes
) -> typ.Tuple[typ.Optional[TypeAbc], ...]:
    type_identifier.read("$_")

    ret = []
    last_was_comma = False

    while True:
        while type_identifier.startswith("_$_"):
            type_identifier.read("_$_")
            if last_was_comma:
                ret.append(None)
            last_was_comma = True
        if type_identifier.startswith("_$"):
            break
        else:
            ret.append(
                TypeAbc.from_type_identifier(
                    type_identifier, reference_resolver, cu_hash
                )
            )
            last_was_comma = False
            if ret[-1] is None:
                # failed to parse => the last character was not a comma but a closing bracket
                type_identifier.insert("_$_")
                ret.pop()
                break

    type_identifier.read("_$")
    return tuple(ret)


def _parse_user_identifier(type_identifier: StringReader) -> str:
    type_identifier.read("$_")

    for match in re.finditer(r"_\$", type_identifier.data):
        if (
            len(type_identifier) < match.start() + 4
            or type_identifier.data[match.start() : match.start() + 4] != "_$$$"
        ):
            name = type_identifier.data[: match.start()]
            type_identifier.read(name + "_$")
            return name.replace("$$$", "$")
    assert False, "Failed to parse user identifier"


class Address(TypeAbc):
    """
    Address type.
    """

    __is_payable: bool

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_address")

        if type_identifier.startswith("_payable"):
            type_identifier.read("_payable")
            self.__is_payable = True
        else:
            self.__is_payable = False

    def abi_type(self) -> str:
        return "address"

    @property
    def is_payable(self) -> bool:
        """
        Returns:
            `True` if the address is payable, `False` otherwise.
        """
        return self.__is_payable


class Bool(TypeAbc):
    """
    Boolean type.
    """

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_bool")

    def abi_type(self) -> str:
        return "bool"


class IntAbc(TypeAbc):
    """
    Base class for [Int][woke.ir.types.Int] and [UInt][woke.ir.types.UInt] types.
    """

    _bits_count: int

    @property
    def bits_count(self) -> int:
        """
        Can only be a multiple of 8, with a minimum of 8 and a maximum of 256.
        Returns:
            Number of bits used to represent this integer.
        """
        return self._bits_count


class Int(IntAbc):
    """
    Signed integer type.
    """

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_int")
        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        number = match.group("number")
        type_identifier.read(number)
        self._bits_count = int(number)

    def abi_type(self) -> str:
        return f"int{self._bits_count}"


class UInt(IntAbc):
    """
    Unsigned integer type.
    """

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_uint")
        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        number = match.group("number")
        type_identifier.read(number)
        self._bits_count = int(number)

    def abi_type(self) -> str:
        return f"uint{self._bits_count}"


class FixedAbc(TypeAbc):
    """
    Base class for [Fixed][woke.ir.types.Fixed] and [UFixed][woke.ir.types.UFixed] types.
    !!! info
        Currently not fully implemented in Solidity.
    """

    _total_bits: int
    _fractional_digits: int

    @property
    def total_bits(self) -> int:
        """
        Returns:
            Total number of bits used to represent this fixed point number.
        """
        return self._total_bits

    @property
    def fractional_digits(self) -> int:
        """
        Returns:
            Number of decimal places available.
        """
        return self._fractional_digits


class Fixed(FixedAbc):
    """
    Signed fixed-point number type as specified by the [Solidity docs](https://docs.soliditylang.org/en/latest/types.html?highlight=fixed#fixed-point-numbers).
    !!! info
        Currently not fully implemented in Solidity.
    """

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_fixed")
        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        total_bits = match.group("number")
        type_identifier.read(total_bits)
        self._total_bits = int(total_bits)

        type_identifier.read("x")

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        fractional_digits = match.group("number")
        type_identifier.read(fractional_digits)
        self._fractional_digits = int(fractional_digits)

    def abi_type(self) -> str:
        raise NotImplementedError


class UFixed(FixedAbc):
    """
    Unsigned fixed point number type as specified by the [Solidity docs](https://docs.soliditylang.org/en/latest/types.html?highlight=ufixed#fixed-point-numbers).
    !!! info
        Currently not fully implemented in Solidity.
    """

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_ufixed")
        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        total_bits = match.group("number")
        type_identifier.read(total_bits)
        self._total_bits = int(total_bits)

        type_identifier.read("x")

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        fractional_digits = match.group("number")
        type_identifier.read(fractional_digits)
        self._fractional_digits = int(fractional_digits)

    def abi_type(self) -> str:
        raise NotImplementedError


class StringLiteral(TypeAbc):
    """
    String literal type.
    !!! warning
        This expression is of the [StringLiteral][woke.ir.types.StringLiteral] type:
        ```solidity
        "Hello, world!"
        ```

        However, this expression is of the [String][woke.ir.types.String] type and contains a child expression of the [StringLiteral][woke.ir.types.StringLiteral] type:
        ```solidity
        string("Hello, world!")
        ```
    """

    __keccak256_hash: bytes

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_stringliteral_")
        match = HEX_RE.match(type_identifier.data)
        assert match is not None
        hex = match.group("hex")
        type_identifier.read(hex)
        self.__keccak256_hash = bytes.fromhex(hex)

    def abi_type(self) -> str:
        raise NotImplementedError

    @property
    def keccak256_hash(self) -> bytes:
        """
        Returns:
            Keccak256 hash of the string literal.
        """
        return self.__keccak256_hash


class String(TypeAbc):
    """
    String type.
    """

    __data_location: DataLocation
    __is_pointer: bool
    __is_slice: bool

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_string")
        if type_identifier.startswith("_storage"):
            self.__data_location = DataLocation.STORAGE
            type_identifier.read("_storage")
        elif type_identifier.startswith("_memory"):
            self.__data_location = DataLocation.MEMORY
            type_identifier.read("_memory")
        elif type_identifier.startswith("_calldata"):
            self.__data_location = DataLocation.CALLDATA
            type_identifier.read("_calldata")
        else:
            assert False, f"Unexpected string type data location {type_identifier}"

        if type_identifier.startswith("_ptr"):
            self.__is_pointer = True
            type_identifier.read("_ptr")
        else:
            self.__is_pointer = False

        if type_identifier.startswith("_slice"):
            self.__is_slice = True
            type_identifier.read("_slice")
        else:
            self.__is_slice = False

    def abi_type(self) -> str:
        return "string"

    @property
    def data_location(self) -> DataLocation:
        """
        Can be either [CALLDATA][woke.ir.enums.DataLocation.CALLDATA], [MEMORY][woke.ir.enums.DataLocation.MEMORY] or [STORAGE][woke.ir.enums.DataLocation.STORAGE]
        Returns:
            Data location of the string expression.
        """
        return self.__data_location

    @property
    def is_pointer(self) -> bool:
        """
        Storage references can be pointers or bound references. In general, local variables are of
        pointer type, state variables are bound references. Assignments to pointers or deleting
        them will not modify storage (that will only change the pointer). Assignment from
        non-storage objects to a variable of storage pointer type is not possible.

        For anything other than [STORAGE][woke.ir.enums.DataLocation.STORAGE], this always returns `True` because assignments
        never change the contents of the original value.

        Returns:
            Whether the string expression is a pointer to storage.
        """
        return self.__is_pointer

    @property
    def is_slice(self) -> bool:
        """
        !!! example
            ```solidity
            function foo(string calldata s) public pure {
                s[0:5]; // s[0:5] is a string slice
            }
            ```
        Returns:
            Whether this is a slice of a string expression.
        """
        return self.__is_slice


class Bytes(TypeAbc):
    """
    Bytes type.
    """

    __data_location: DataLocation
    __is_pointer: bool
    __is_slice: bool

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_bytes")
        if type_identifier.startswith("_storage"):
            self.__data_location = DataLocation.STORAGE
            type_identifier.read("_storage")
        elif type_identifier.startswith("_memory"):
            self.__data_location = DataLocation.MEMORY
            type_identifier.read("_memory")
        elif type_identifier.startswith("_calldata"):
            self.__data_location = DataLocation.CALLDATA
            type_identifier.read("_calldata")
        else:
            assert False, f"Unexpected string type data location {type_identifier}"

        if type_identifier.startswith("_ptr"):
            self.__is_pointer = True
            type_identifier.read("_ptr")
        else:
            self.__is_pointer = False

        if type_identifier.startswith("_slice"):
            self.__is_slice = True
            type_identifier.read("_slice")
        else:
            self.__is_slice = False

    def abi_type(self) -> str:
        return "bytes"

    @property
    def data_location(self) -> DataLocation:
        """
        Can be either [CALLDATA][woke.ir.enums.DataLocation.CALLDATA], [MEMORY][woke.ir.enums.DataLocation.MEMORY] or [STORAGE][woke.ir.enums.DataLocation.STORAGE]
        Returns:
            Data location of the bytes expression.
        """
        return self.__data_location

    @property
    def is_pointer(self) -> bool:
        """
        Storage references can be pointers or bound references. In general, local variables are of
        pointer type, state variables are bound references. Assignments to pointers or deleting
        them will not modify storage (that will only change the pointer). Assignment from
        non-storage objects to a variable of storage pointer type is not possible.

        For anything other than [STORAGE][woke.ir.enums.DataLocation.STORAGE], this always returns `True` because assignments
        never change the contents of the original value.

        Returns:
            Whether the bytes expression is a pointer to storage.
        """
        return self.__is_pointer

    @property
    def is_slice(self) -> bool:
        """
        !!! example
            ```solidity
            function foo(bytes calldata b) public pure {
                b[0:5]; // s[0:5] is a bytes slice
            }
            ```
        Returns:
            Whether this is a slice of a bytes expression.
        """
        return self.__is_slice


class FixedBytes(TypeAbc):
    """
    Fixed-size byte array type.
    """

    __bytes_count: int

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_bytes")
        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None
        number = match.group("number")
        type_identifier.read(number)
        self.__bytes_count = int(number)

    def abi_type(self) -> str:
        return f"bytes{self.__bytes_count}"

    @property
    def bytes_count(self) -> int:
        """
        Is at least 1 and at most 32.
        Returns:
            Number of bytes used to represent this fixed-size byte array.
        """
        return self.__bytes_count


class Function(TypeAbc):
    """
    Function type.

    !!! warning
        Given the following function:
        ```solidity
        function foo(uint a, uint b) public pure returns(uint, uint) {
            return (a + b, a - b);
        }
        ```
        and the following call:
        ```solidity
        foo(1, 2);
        ```
        the type of `foo` is [Function][woke.ir.types.Function], but the type of `:::solidity foo(1, 2)` is [Tuple][woke.ir.types.Tuple].

    """

    __kind: FunctionTypeKind
    __state_mutability: StateMutability
    __parameters: typ.Tuple[TypeAbc, ...]
    __return_parameters: typ.Tuple[TypeAbc, ...]
    __gas_set: bool
    __value_set: bool
    __salt_set: bool
    __attached_to: typ.Optional[typ.Tuple[TypeAbc, ...]]

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_function_")

        matched = []
        for kind in FunctionTypeKind:
            if type_identifier.startswith(kind):
                matched.append(kind)
        assert len(matched) >= 1, f"Unexpected function kind {type_identifier}"
        self.__kind = FunctionTypeKind(max(matched, key=len))
        type_identifier.read(self.__kind)

        if type_identifier.startswith("_payable"):
            self.__state_mutability = StateMutability.PAYABLE
            type_identifier.read("_payable")
        elif type_identifier.startswith("_pure"):
            self.__state_mutability = StateMutability.PURE
            type_identifier.read("_pure")
        elif type_identifier.startswith("_nonpayable"):
            self.__state_mutability = StateMutability.NONPAYABLE
            type_identifier.read("_nonpayable")
        elif type_identifier.startswith("_view"):
            self.__state_mutability = StateMutability.VIEW
            type_identifier.read("_view")
        else:
            assert False, f"Unexpected function state mutability {type_identifier}"

        parameters = _parse_list(type_identifier, reference_resolver, cu_hash)
        assert not any(param is None for param in parameters)
        self.__parameters = parameters  # type: ignore

        type_identifier.read("returns")
        return_parameters = _parse_list(type_identifier, reference_resolver, cu_hash)
        assert not any(param is None for param in return_parameters)
        self.__return_parameters = return_parameters  # type: ignore

        if type_identifier.startswith("gas"):
            self.__gas_set = True
            type_identifier.read("gas")
        else:
            self.__gas_set = False

        if type_identifier.startswith("value"):
            self.__value_set = True
            type_identifier.read("value")
        else:
            self.__value_set = False

        if type_identifier.startswith("salt"):
            self.__salt_set = True
            type_identifier.read("salt")
        else:
            self.__salt_set = False

        if type_identifier.startswith("bound_to"):
            type_identifier.read("bound_to")
            bound_to = _parse_list(type_identifier, reference_resolver, cu_hash)
            assert not any(param is None for param in bound_to)
            self.__attached_to = bound_to  # type: ignore
        elif type_identifier.startswith(
            "attached_to"
        ):  # bound_to was renamed to attached_to in 0.8.18
            type_identifier.read("attached_to")
            attached_to = _parse_list(type_identifier, reference_resolver, cu_hash)
            assert not any(param is None for param in attached_to)
            self.__attached_to = attached_to  # type: ignore
        else:
            self.__attached_to = None

    def abi_type(self) -> str:
        return "function"

    @property
    def kind(self) -> FunctionTypeKind:
        """
        Returns:
            Kind of the function type.
        """
        return self.__kind

    @property
    def state_mutability(self) -> StateMutability:
        """
        Returns:
            State mutability of the function type.
        """
        return self.__state_mutability

    @property
    def parameters(self) -> typ.Tuple[TypeAbc, ...]:
        """
        Returns:
            Expression types of the parameters of the function type.
        """
        return self.__parameters

    @property
    def return_parameters(self) -> typ.Tuple[TypeAbc, ...]:
        """
        Returns:
            Expression types of the return parameters of the function type.
        """
        return self.__return_parameters

    @property
    def gas_set(self) -> bool:
        """
        !!! example
            In the case of the old syntax (deprecated), the `gas` [MemberAccess][woke.ir.expressions.member_access.MemberAccess] expression is of the [Function][woke.ir.types.Function] type which returns a [Function][woke.ir.types.Function] with `gas_set` set to `True`.
            ```solidity
            foo.gas(10)(1, 2);
            ```

            In the case of the new syntax, the `{gas: 10}` [FunctionCallOptions][woke.ir.expressions.function_call_options.FunctionCallOptions] expression is of the [Function][woke.ir.types.Function] type with `gas_set` set to `True`.
            ```solidity
            foo{gas: 10}(1, 2);
            ```
        Returns:
            `True` if the gas is set in the function type.
        """
        return self.__gas_set

    @property
    def value_set(self) -> bool:
        """
        !!! example
            In the case of the old syntax (deprecated), the `value` [MemberAccess][woke.ir.expressions.member_access.MemberAccess] expression is of the [Function][woke.ir.types.Function] type which returns a [Function][woke.ir.types.Function] with `value_set` set to `True`.
            ```solidity
            foo.value(1)(1, 2);
            ```

            In the case of the new syntax, the `{value: 1}` [FunctionCallOptions][woke.ir.expressions.function_call_options.FunctionCallOptions] expression is of the [Function][woke.ir.types.Function] type with `value_set` set to `True`.
            ```solidity
            foo{value: 1}(1, 2);
            ```
        Returns:
            `True` if the value is set in the function type.
        """
        return self.__value_set

    @property
    def salt_set(self) -> bool:
        """
        !!! example
            In the following example, the `{salt: salt}` [FunctionCallOptions][woke.ir.expressions.function_call_options.FunctionCallOptions] expression is of the [Function][woke.ir.types.Function] type with `salt_set` set to `True`.
            ```solidity
            new Foo{salt: salt}();
            ```
        Returns:
            `True` if the salt is set in the function type.
        """
        return self.__salt_set

    @property
    def attached_to(self) -> typ.Optional[typ.Tuple[TypeAbc, ...]]:
        """
        A function type can be attached to a type using the [UsingForDirective][woke.ir.meta.using_for_directive.UsingForDirective] or internally in the case of a Solidity global symbol.
        !!! example
            In the following example, the `add` [MemberAccess][woke.ir.expressions.member_access.MemberAccess] expression on line 9 is of the [Function][woke.ir.types.Function] type and is attached to the [UInt][woke.ir.types.UInt] type.
            ```solidity linenums="1"
            function add(uint a, uint b) pure returns (uint) {
                return a + b;
            }

            using {add} for uint;

            contract Foo {
                function bar(uint x) public pure returns(uint) {
                    return x.add(1);
                }
            }
            ```

            In this example, the `push` [MemberAccess][woke.ir.expressions.member_access.MemberAccess] expression on line 9 is of the [Function][woke.ir.types.Function] type and is attached to the [Array][woke.ir.types.Array] type.
            ```solidity
            arr.push(1);
            ```

        Returns:
            Type to which the function is attached to.
        """
        return self.__attached_to


class Tuple(TypeAbc):
    """
    Tuple type.
    """

    __components: typ.Tuple[typ.Optional[TypeAbc], ...]

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_tuple")
        self.__components = _parse_list(type_identifier, reference_resolver, cu_hash)

    def abi_type(self) -> str:
        if any(component is None for component in self.__components):
            raise NotImplementedError
        return (
            "("
            + ",".join(
                component.abi_type()  # pyright: ignore reportOptionalMemberAccess
                for component in self.__components
            )
            + ")"
        )

    @property
    def components(self) -> typ.Tuple[typ.Optional[TypeAbc], ...]:
        """
        A component type can be `None` in the case of a tuple with a missing component.
        !!! example
            In the following example, the `(success, )` expression is of the [Tuple][woke.ir.types.Tuple] type with the components of the type [Bool][woke.ir.types.Bool] and `None`.
            ```solidity
            bool success;
            (success, ) = address(addr).call{value: 1}("");
            ```

        Returns:
            Expression types of the components of the tuple type.
        """
        return self.__components


class Type(TypeAbc):
    """
    Type type. As opposed to other types, this type describes the type of a type, not the type of an instance of a type.
    """

    __actual_type: TypeAbc

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_type")
        actual_type = _parse_list(type_identifier, reference_resolver, cu_hash)
        assert len(actual_type) == 1 and actual_type[0] is not None
        self.__actual_type = actual_type[0]

    def abi_type(self) -> str:
        raise NotImplementedError

    @property
    def actual_type(self) -> TypeAbc:
        """
        !!! example
            `payable` in the following example is of the [Type][woke.ir.types.Type] type with the [Address][woke.ir.types.Address] actual type.
            ```solidity
            payable(addr);
            ```

            `super` in the following example is of the [Type][woke.ir.types.Type] type with the [Contract][woke.ir.types.Contract] actual type.
            ```solidity
            super.foo();
            ```

            `string` in the following example is of the [Type][woke.ir.types.Type] type with the [String][woke.ir.types.String] actual type.
            ```solidity
            string.concat("foo", "bar");
            ```

            `Foo` in the following example on line 4 is of the [Type][woke.ir.types.Type] type with the [Enum][woke.ir.types.Enum] actual type.
            ```solidity linenums="1"
            enum Foo { A, B }

            function bar() pure returns (Foo) {
                return Foo.A;
            }
            ```

        Returns:
            Actual type of the type type.
        """
        return self.__actual_type


class Rational(TypeAbc):
    """
    Rational type. Represents the type of constants or expressions with constants.

    !!! example
        `:::solidity 1`, `:::solidity 0x1234`, `:::solidity 1e18`, `:::solidity 1 * 2 / 3` are all of the [Rational][woke.ir.types.Rational] type.
    """

    __numerator: int
    __denominator: int

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_rational_")

        if type_identifier.startswith("minus_"):
            type_identifier.read("minus_")
            self.__numerator = -1
        else:
            self.__numerator = 1

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid rational"
        number = match.group("number")
        type_identifier.read(number)
        self.__numerator *= int(number)

        type_identifier.read("_by_")

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid rational"
        number = match.group("number")
        type_identifier.read(number)
        self.__denominator = int(number)

    def abi_type(self) -> str:
        raise NotImplementedError

    @property
    def numerator(self) -> int:
        """
        If the rational is negative, the numerator will be negative.
        Returns:
            Numerator of the rational number.
        """
        return self.__numerator

    @property
    def denominator(self) -> int:
        """
        Returns:
            Denominator of the rational number.
        """
        return self.__denominator


class Modifier(TypeAbc):
    """
    Modifier type.
    """

    __parameters: typ.Tuple[TypeAbc, ...]

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_modifier")
        parameters = _parse_list(type_identifier, reference_resolver, cu_hash)
        assert not any(param is None for param in parameters)
        self.__parameters = parameters  # type: ignore

    def abi_type(self) -> str:
        raise NotImplementedError

    @property
    def parameters(self) -> typ.Tuple[TypeAbc, ...]:
        """
        Returns:
            Expression types of the parameters of the modifier.
        """
        return self.__parameters


class Array(TypeAbc):
    """
    Array type.
    """

    __base_type: TypeAbc
    __length: typ.Optional[int]
    __data_location: DataLocation
    __is_pointer: bool
    __is_slice: bool

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_array")
        base_type = _parse_list(type_identifier, reference_resolver, cu_hash)
        assert (
            len(base_type) == 1 and base_type[0] is not None
        ), f"Unexpected array base type {type_identifier}"
        self.__base_type = base_type[0]

        if type_identifier.startswith("dyn"):
            self.__length = None
            type_identifier.read("dyn")
        else:
            match = NUMBER_RE.match(type_identifier.data)
            assert match is not None, f"{type_identifier} is not a valid array length"
            self.__length = int(match.group("number"))
            type_identifier.read(match.group("number"))

        if type_identifier.startswith("_storage"):
            self.__data_location = DataLocation.STORAGE
            type_identifier.read("_storage")
        elif type_identifier.startswith("_memory"):
            self.__data_location = DataLocation.MEMORY
            type_identifier.read("_memory")
        elif type_identifier.startswith("_calldata"):
            self.__data_location = DataLocation.CALLDATA
            type_identifier.read("_calldata")
        else:
            assert False, f"Unexpected array type data location {type_identifier}"

        if type_identifier.startswith("_ptr"):
            self.__is_pointer = True
            type_identifier.read("_ptr")
        else:
            self.__is_pointer = False

        if type_identifier.startswith("_slice"):
            self.__is_slice = True
            type_identifier.read("_slice")
        else:
            self.__is_slice = False

    def abi_type(self) -> str:
        if self.length is not None:
            return f"{self.base_type.abi_type()}[{self.length}]"
        else:
            return f"{self.base_type.abi_type()}[]"

    @property
    def base_type(self) -> TypeAbc:
        """
        Returns:
            Base type of the array.
        """
        return self.__base_type

    @property
    def length(self) -> typ.Optional[int]:
        """
        Returns:
            Length of the array. `None` if the array is dynamic (not fixed size).
        """
        return self.__length

    @property
    def data_location(self) -> DataLocation:
        """
        Returns:
            Data location of the array.
        """
        return self.__data_location

    @property
    def is_pointer(self) -> bool:
        """
        Storage references can be pointers or bound references. In general, local variables are of
        pointer type, state variables are bound references. Assignments to pointers or deleting
        them will not modify storage (that will only change the pointer). Assignment from
        non-storage objects to a variable of storage pointer type is not possible.

        For anything other than [STORAGE][woke.ir.enums.DataLocation.STORAGE], this always returns `True` because assignments
        never change the contents of the original value.

        Returns:
            Whether the array expression is a pointer to storage.
        """
        return self.__is_pointer

    @property
    def is_slice(self) -> bool:
        """
        !!! example
            ```solidity
            function foo(uint[] calldata arr) public pure {
                arr[0:5]; // arr[0:5] is an array slice
            }
            ```
        Returns:
            Whether this is a slice of an array expression.
        """
        return self.__is_slice


class Mapping(TypeAbc):
    """
    Mapping type.
    """

    __key_type: TypeAbc
    __value_type: TypeAbc

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_mapping")
        key_value = _parse_list(type_identifier, reference_resolver, cu_hash)
        assert len(key_value) == 2, f"{type_identifier} is not a valid mapping"
        assert key_value[0] is not None, f"{type_identifier} is not a valid mapping"
        assert key_value[1] is not None, f"{type_identifier} is not a valid mapping"
        self.__key_type = key_value[0]
        self.__value_type = key_value[1]

    def abi_type(self) -> str:
        raise NotImplementedError

    @property
    def key_type(self) -> TypeAbc:
        """
        Returns:
            Key type of the mapping.
        """
        return self.__key_type

    @property
    def value_type(self) -> TypeAbc:
        """
        Returns:
            Value type of the mapping.
        """
        return self.__value_type


class Contract(TypeAbc):
    __is_super: bool
    __name: str
    __ast_id: AstNodeId
    __reference_resolver: ReferenceResolver
    __cu_hash: bytes

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        if type_identifier.startswith("t_contract"):
            self.__is_super = False
            type_identifier.read("t_contract")
        elif type_identifier.startswith("t_super"):
            self.__is_super = True
            type_identifier.read("t_super")
        else:
            assert False, f"Unexpected contract type {type_identifier}"
        self.__name = _parse_user_identifier(type_identifier)

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid contract"
        self.__ast_id = AstNodeId(int(match.group("number")))
        type_identifier.read(match.group("number"))

        self.__reference_resolver = reference_resolver
        self.__cu_hash = cu_hash

    def abi_type(self) -> str:
        return "address"

    @property
    def is_super(self) -> bool:
        """
        !!! warning
            Until 0.7.6, the `super` keyword ([Identifier][woke.ir.expressions.identifier.Identifier]) was of the [Contract][woke.ir.types.Contract] type with `is_super` set to `True`.
            Since 0.8.0, the `super` keyword is of the [Type][woke.ir.types.Type] type with [Contract][woke.ir.types.Contract] as the `actual_type` and `is_super` set to `True`.

        !!! warning
            When this is `True`, the `name` and `ir_node` properties refer to the current contract, not the base contract.

        !!! example
            The `name` and `ir_node` properties of the [Contract][woke.ir.types.Contract] type of the `super` expression in the following example refer to the `Foo` contract, not the `Bar` contract.
            ```solidity
            contract Foo is Bar {
                function foo() public {
                    super.foo();
                }
            }
            ```

        Returns:
            `True` if the expression is the `super` keyword.
        """
        return self.__is_super

    @property
    def name(self) -> str:
        """
        Returns:
            Name of the contract.
        """
        return self.__name

    @property
    def ir_node(self) -> ContractDefinition:
        """
        Returns:
            Contract definition IR node.
        """
        from woke.ir.declarations.contract_definition import ContractDefinition

        node = self.__reference_resolver.resolve_node(self.__ast_id, self.__cu_hash)
        assert isinstance(node, ContractDefinition)
        return node


class Struct(TypeAbc):
    """
    Struct type.
    """

    __name: str
    __ast_id: AstNodeId
    __data_location: DataLocation
    __is_pointer: bool
    __reference_resolver: ReferenceResolver
    __cu_hash: bytes

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_struct")
        self.__name = _parse_user_identifier(type_identifier)

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid struct"
        self.__ast_id = AstNodeId(int(match.group("number")))
        type_identifier.read(match.group("number"))

        if type_identifier.startswith("_storage"):
            self.__data_location = DataLocation.STORAGE
            type_identifier.read("_storage")
        elif type_identifier.startswith("_memory"):
            self.__data_location = DataLocation.MEMORY
            type_identifier.read("_memory")
        elif type_identifier.startswith("_calldata"):
            self.__data_location = DataLocation.CALLDATA
            type_identifier.read("_calldata")
        else:
            assert False, f"Unexpected array type data location {type_identifier}"

        if type_identifier.startswith("_ptr"):
            self.__is_pointer = True
            type_identifier.read("_ptr")
        else:
            self.__is_pointer = False

        self.__reference_resolver = reference_resolver
        self.__cu_hash = cu_hash

    def abi_type(self) -> str:
        return (
            "("
            + ",".join([member.type.abi_type() for member in self.ir_node.members])
            + ")"
        )

    @property
    def name(self) -> str:
        """
        Returns:
            Name of the struct.
        """
        return self.__name

    @property
    def data_location(self) -> DataLocation:
        """
        Returns:
            Data location of the struct.
        """
        return self.__data_location

    @property
    def is_pointer(self) -> bool:
        """
        Storage references can be pointers or bound references. In general, local variables are of
        pointer type, state variables are bound references. Assignments to pointers or deleting
        them will not modify storage (that will only change the pointer). Assignment from
        non-storage objects to a variable of storage pointer type is not possible.

        For anything other than [STORAGE][woke.ir.enums.DataLocation.STORAGE], this always returns `True` because assignments
        never change the contents of the original value.

        Returns:
            Whether the struct expression is a pointer to storage.
        """
        return self.__is_pointer

    @property
    def ir_node(self) -> StructDefinition:
        """
        Returns:
            Struct definition IR node.
        """
        from woke.ir.declarations.struct_definition import StructDefinition

        node = self.__reference_resolver.resolve_node(self.__ast_id, self.__cu_hash)
        assert isinstance(node, StructDefinition)
        return node


class Enum(TypeAbc):
    """
    Enum type.

    !!! warning
        Enum values are of the [Enum][woke.ir.types.Enum] type and enums are of the [Type][woke.ir.types.Type] type with [Enum][woke.ir.types.Enum] as the [actual_type][woke.ir.types.Type.actual_type].
    """

    __name: str
    __ast_id: AstNodeId
    __reference_resolver: ReferenceResolver
    __cu_hash: bytes

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_enum")
        self.__name = _parse_user_identifier(type_identifier)

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid enum"
        self.__ast_id = AstNodeId(int(match.group("number")))
        type_identifier.read(match.group("number"))

        self.__reference_resolver = reference_resolver
        self.__cu_hash = cu_hash

    def abi_type(self) -> str:
        return "uint8"

    @property
    def name(self) -> str:
        """
        Returns:
            Name of the enum.
        """
        return self.__name

    @property
    def ir_node(self) -> EnumDefinition:
        """
        Returns:
            Enum definition IR node.
        """
        from woke.ir.declarations.enum_definition import EnumDefinition

        node = self.__reference_resolver.resolve_node(self.__ast_id, self.__cu_hash)
        assert isinstance(node, EnumDefinition)
        return node


class Magic(TypeAbc):
    """
    Magic type represents Solidity language built-in objects.
    """

    __kind: MagicTypeKind
    __meta_argument_type: typ.Optional[TypeAbc]

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_magic_")

        matched = False
        for kind in MagicTypeKind:
            if type_identifier.startswith(kind):
                self.__kind = MagicTypeKind(kind)
                type_identifier.read(kind)
                matched = True
                break
        assert matched, f"Unexpected magic kind {type_identifier}"

        if self.__kind == MagicTypeKind.META_TYPE:
            type_identifier.read("_")
            meta_argument_type = TypeAbc.from_type_identifier(
                type_identifier, reference_resolver, cu_hash
            )
            assert meta_argument_type is not None
            self.__meta_argument_type = meta_argument_type
        else:
            self.__meta_argument_type = None

    def abi_type(self) -> str:
        raise NotImplementedError

    @property
    def kind(self) -> MagicTypeKind:
        """
        Returns:
            Kind of the magic type.
        """
        return self.__kind

    @property
    def meta_argument_type(self) -> typ.Optional[TypeAbc]:
        """
        Is only set for [MagicTypeKind.META_TYPE][woke.ir.enums.MagicTypeKind.META_TYPE] kind.
        !!! example
            [Contract][woke.ir.types.Contract] in `:::solidity type(IERC20)`, [UInt][woke.ir.types.UInt] in `:::solidity type(uint)`.

        Returns:
            Type of the meta expression argument.
        """
        return self.__meta_argument_type


class UserDefinedValueType(TypeAbc):
    __name: str
    __ast_id: AstNodeId
    __reference_resolver: ReferenceResolver
    __cu_hash: bytes

    def __init__(
        self,
        type_identifier: StringReader,
        reference_resolver: ReferenceResolver,
        cu_hash: bytes,
    ):
        type_identifier.read("t_userDefinedValueType")
        self.__name = _parse_user_identifier(type_identifier)

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid enum"
        self.__ast_id = AstNodeId(int(match.group("number")))
        type_identifier.read(match.group("number"))

        self.__reference_resolver = reference_resolver
        self.__cu_hash = cu_hash

    def abi_type(self) -> str:
        return self.ir_node.underlying_type.type.abi_type()

    @property
    def name(self) -> str:
        """
        Returns:
            Name of the user defined value type.
        """
        return self.__name

    @property
    def ir_node(self) -> UserDefinedValueTypeDefinition:
        """
        Returns:
            User defined value type definition IR node.
        """
        from woke.ir.declarations.user_defined_value_type_definition import (
            UserDefinedValueTypeDefinition,
        )

        node = self.__reference_resolver.resolve_node(self.__ast_id, self.__cu_hash)
        assert isinstance(node, UserDefinedValueTypeDefinition)
        return node


class Module(TypeAbc):
    """
    Module type.
    !!! note
        It is probably currently not possible to create an expression of this type.
    """

    __source_unit_id: int

    def __init__(self, type_identifier: StringReader):
        type_identifier.read("t_module_")

        match = NUMBER_RE.match(type_identifier.data)
        assert match is not None, f"{type_identifier} is not a valid module"
        self.__ast_id = AstNodeId(int(match.group("number")))
        type_identifier.read(match.group("number"))

        self.__source_unit_id = int(match.group("number"))

    def abi_type(self) -> str:
        raise NotImplementedError



import enum
from typing import Any, Optional, Union

from .lsp_data_model import LspModel
from .methods import RequestMethodEnum


class Message(LspModel):
    jsonrpc: str


class RequestMessage(Message):
    id: Union[int, str]
    """
    The request id.
    """
    method: RequestMethodEnum
    """
    The method to be invoked.
    """
    params: Optional[Any]


class ResponseError(LspModel):
    code: int
    """
    A number indicating the error type that occurred.
    """
    message: str
    """
    A string providing a short description of the error.
    """
    data: Optional[Any]
    """
    A primitive or structured value that contains additional
    information about the error. Can be omitted.
    """


class ResponseMessage(Message):
    id: Union[int, str, None]
    """
    The request id.
    """
    result: Optional[Any]
    """
    The result of a request. This member is REQUIRED on success.
    This member MUST NOT exist if there was an error invoking the method.
    """
    error: Optional[ResponseError]
    """
    The error object in case a request fails.
    """


class NotificationMessage(Message):
    method: str
    """
    The method to be invoked.
    """
    params: Optional[Any]
    """
    The notification's params.
    """


class CancelParams(LspModel):
    id: Union[int, str]
    """
    The request id to cancel.
    """


class ErrorCodes(enum.IntEnum):
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    ServerNotInitialized = -32002
    UnknownErrorCode = -32001
    RequestFailed = -32803
    ServerCancelled = -32802
    ContentModified = -32801
    RequestCancelled = -32800
    # jsonrpcReservedErrorRangeStart = -32099
    # jsonrpcReservedErrorRangeEnd = -32000
    # lspReservedErrorRangeStart = -32899
    # lspReservedErrorRangeEnd = -32800



from __future__ import annotations

from typing import TYPE_CHECKING, Union

from woke.ir.abc import SolidityAbc
from woke.ir.ast import SolcStructuredDocumentation
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from ..declarations.contract_definition import ContractDefinition
    from ..declarations.error_definition import ErrorDefinition
    from ..declarations.event_definition import EventDefinition
    from ..declarations.function_definition import FunctionDefinition
    from ..declarations.modifier_definition import ModifierDefinition
    from ..declarations.variable_declaration import VariableDeclaration


class StructuredDocumentation(SolidityAbc):
    """
    !!! example
        Lines 1-4 in the following example:
        ```solidity linenums="1"
        /// @title A simulator for trees
        /// @author John
        /// @notice You can use this contract for only the most basic simulation
        /// @dev All function calls are currently implemented without side effects
        contract Tree {
            function multiply(uint a) public pure returns(uint) {
                return a * 7;
            }
        }
        ```
    """

    _ast_node: SolcStructuredDocumentation
    _parent: Union[
        ContractDefinition,
        ErrorDefinition,
        EventDefinition,
        FunctionDefinition,
        ModifierDefinition,
        VariableDeclaration,
    ]

    _text: str

    def __init__(
        self,
        init: IrInitTuple,
        structured_documentation: SolcStructuredDocumentation,
        parent: SolidityAbc,
    ):
        super().__init__(init, structured_documentation, parent)
        self._text = structured_documentation.text

    @property
    def parent(
        self,
    ) -> Union[
        ContractDefinition,
        ErrorDefinition,
        EventDefinition,
        FunctionDefinition,
        ModifierDefinition,
        VariableDeclaration,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def text(self) -> str:
        """
        Does not include the leading `///` or `/**` and trailing `*/`.
        Returns:
            [NatSpec](https://docs.soliditylang.org/en/latest/natspec-format.html) documentation string.
        """
        return self._text



from functools import wraps

from .context_managers import recursion_guard


def return_on_recursion(default):
    arguments = set()

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if len(kwargs) > 0:
                raise ValueError("kwargs not supported")
            if args in arguments:
                return default
            with recursion_guard(arguments, *args):
                return func(*args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

from typing import Deque, List, Set

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


class MsgValueNonpayableFunctionDetector(Detector):
    _detections: List[DetectorResult]

    def __init__(self):
        self._detections = []

    def detect(self) -> List[DetectorResult]:
        return self._detections

    def visit_member_access(self, node):
        from collections import deque
        from itertools import chain

        if node.referenced_declaration != ir.enums.GlobalSymbolsEnum.MSG_VALUE:
            return

        func = node
        while func is not None:
            if isinstance(func, ir.FunctionDefinition):
                break
            func = func.parent

        if func is None:
            return

        if func.state_mutability == ir.enums.StateMutability.PAYABLE:
            return

        visited: Set[ir.FunctionDefinition] = {func}
        queue: Deque[ir.FunctionDefinition] = deque([func])

        while len(queue) > 0:
            func = queue.pop()
            for f in chain(func.base_functions, func.child_functions):
                if isinstance(f, ir.FunctionDefinition) and f not in visited:
                    visited.add(f)
                    queue.append(f)

            for ref in func.references:
                if isinstance(ref, ir.IdentifierPathPart):
                    func_call = ref.underlying_node
                elif isinstance(ref, ir.ExternalReference):
                    func_call = ref.inline_assembly
                else:
                    func_call = ref
                while func_call is not None:
                    if (
                        isinstance(func_call, ir.FunctionCall)
                        and func_call.function_called == func
                    ):
                        break
                    func_call = func_call.parent

                if func_call is None:
                    continue

                other_func = func_call
                while other_func is not None:
                    if isinstance(other_func, ir.FunctionDefinition):
                        break
                    other_func = other_func.parent

                if other_func is None or other_func in visited:
                    continue

                if other_func.state_mutability == ir.enums.StateMutability.PAYABLE:
                    return

                queue.append(other_func)

        self._detections.append(
            DetectorResult(
                Detection(
                    node,
                    "`msg.value` used in a function that is never called from a payable function",
                ),
                impact=DetectionImpact.MEDIUM,
                confidence=DetectionConfidence.HIGH,
            )
        )

    @detector.command("msg-value-nonpayable-function")
    def cli(self):
        pass



from typing import Optional, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcLiteral
from woke.ir.enums import LiteralKind, ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.utils import IrInitTuple


class Literal(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcLiteral
    _parent: SolidityAbc  # TODO: make this more specific

    _hex_value: str
    _kind: LiteralKind
    _subdenomination: Optional[str]
    _value: Optional[str]

    def __init__(self, init: IrInitTuple, literal: SolcLiteral, parent: SolidityAbc):
        super().__init__(init, literal, parent)
        self._hex_value = literal.hex_value
        self._kind = literal.kind
        self._subdenomination = literal.subdenomination
        self._value = literal.value

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def hex_value(self) -> str:
        return self._hex_value

    @property
    def kind(self) -> LiteralKind:
        return self._kind

    @property
    def subdenomination(self) -> Optional[str]:
        return self._subdenomination

    @property
    def value(self) -> Optional[str]:
        return self._value

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return set()



from pathlib import Path
from typing import Dict, FrozenSet, List, NamedTuple, Optional

import pydantic
from intervaltree import IntervalTree
from pydantic import BaseModel, Extra

from woke.compiler.solc_frontend import SolcInputSettings, SolcOutputError
from woke.core.solidity_version import SolidityVersion
from woke.ir import SourceUnit
from woke.ir.reference_resolver import ReferenceResolver


class BuildInfoModel(BaseModel):
    class Config:
        extra = Extra.allow
        allow_mutation = False
        arbitrary_types_allowed = True
        json_encoders = {SolidityVersion: str, bytes: lambda b: b.hex()}


class CompilationUnitBuildInfo(BuildInfoModel):
    errors: List[SolcOutputError]


# workaround for pydantic bytes JSON encode bug: https://github.com/pydantic/pydantic/issues/3756
class HexBytes(bytes):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, bytes):
            return v
        elif isinstance(v, bytearray):
            return bytes(v)
        elif isinstance(v, str):
            return bytes.fromhex(v)
        raise pydantic.errors.BytesError()


class SourceUnitInfo(NamedTuple):
    fs_path: Path
    blake2b_hash: HexBytes


class ProjectBuildInfo(BuildInfoModel):
    compilation_units: Dict[str, CompilationUnitBuildInfo]
    source_units_info: Dict[str, SourceUnitInfo]
    allow_paths: FrozenSet[Path]
    ignore_paths: FrozenSet[Path]
    include_paths: FrozenSet[Path]
    settings: SolcInputSettings
    target_solidity_version: Optional[SolidityVersion]
    woke_version: str
    incremental: bool


class ProjectBuild(NamedTuple):
    interval_trees: Dict[Path, IntervalTree]
    reference_resolver: ReferenceResolver
    source_units: Dict[Path, SourceUnit]



from functools import lru_cache
from typing import Iterator, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcConditional
from woke.ir.enums import ModifiesStateFlag
from woke.ir.utils import IrInitTuple

from .abc import ExpressionAbc


class Conditional(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcConditional
    _parent: SolidityAbc  # TODO: make this more specific

    _condition: ExpressionAbc
    _false_expression: ExpressionAbc
    _true_expression: ExpressionAbc

    def __init__(
        self, init: IrInitTuple, conditional: SolcConditional, parent: SolidityAbc
    ):
        super().__init__(init, conditional, parent)
        self._condition = ExpressionAbc.from_ast(init, conditional.condition, self)
        self._false_expression = ExpressionAbc.from_ast(
            init, conditional.false_expression, self
        )
        self._true_expression = ExpressionAbc.from_ast(
            init, conditional.true_expression, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._condition
        yield from self._false_expression
        yield from self._true_expression

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def condition(self) -> ExpressionAbc:
        return self._condition

    @property
    def false_expression(self) -> ExpressionAbc:
        return self._false_expression

    @property
    def true_expression(self) -> ExpressionAbc:
        return self._true_expression

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        return (
            self.true_expression.is_ref_to_state_variable
            or self.false_expression.is_ref_to_state_variable
        )

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return (
            self.condition.modifies_state
            | self.true_expression.modifies_state
            | self.false_expression.modifies_state
        )



from enum import IntEnum
from typing import Any, Dict, List, NewType, Optional, Tuple, Union

from woke.utils import StrEnum

from .lsp_data_model import LspModel

DocumentUri = NewType("URI", str)
URI = NewType("URI", str)
ChangeAnnotationIdentifier = NewType("ChangeAnnotationIdentifier", str)
ProgressToken = Union[int, str]
TraceValue = NewType("Trace", str)  # NewType(Union["off","message","verbose"], str)


class RegularExpressionsClientCapabilities(LspModel):
    engine: str
    """
    The engine's name.
    """
    version: Optional[str]
    """
    The engine's version.
    """


class Position(LspModel):
    line: int
    """
    Position in a document (zero-based).
    """
    character: int
    """
    Character offset on a line in a document (zero-based).between the `character` and `character + 1`.
    """

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.line, self.character) == (other.line, other.character)
        return NotImplemented

    def __hash__(self):
        return hash((self.line, self.character))


class Range(LspModel):
    start: Position
    """
    The range's start position.
    """
    end: Position
    """
    The range's end position.
    """

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.start, self.end) == (other.start, other.end)
        return NotImplemented

    def __hash__(self):
        return hash((self.start, self.end))


class Location(LspModel):
    """
    Represents a location inside a resource, such as a line inside a text file.
    """

    uri: DocumentUri
    range: Range

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.uri, self.range) == (other.uri, other.range)
        return NotImplemented

    def __hash__(self):
        return hash((self.uri, self.range))


class LocationLink(LspModel):
    origin_selection_range: Optional[Range]
    """
    Span of the origin of this link.
    Used as the underlined span for mouse interaction.
    Defaults to the word range at the mouse position.
    """
    target_uri: str
    """
     The target resource identifier of this link.
    """
    target_range: Range
    """
    The full target range of this link.
    """
    target_selection_range: Range
    """
    The range that should be selected and revealed when this link is being followed,
    e.g the name of a function. Must be contained by the the `target_Range`.
    """


class DiagnosticSeverity(IntEnum):
    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    HINT = 4


class DiagnosticTag(IntEnum):
    UNNECESSARY = 1
    """ 
    Unused or unnecessary code
    """
    DEPRECATED = 2
    """
    Deprecated or obsolete code 
    """


class DiagnosticRelatedInformation(LspModel):
    location: Location
    """
    The location of this related diagnostic information.
    """
    message: str
    """
    The message of this related diagnostic information.
    """

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return (self.location, self.message) == (other.location, other.message)
        return NotImplemented

    def __hash__(self):
        return hash((self.location, self.message))


class CodeDescription(LspModel):
    href: URI
    """
    URI to open with more info.
    """

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.href == other.href
        return NotImplemented

    def __hash__(self):
        return hash(self.href)


class Diagnostic(LspModel):
    range: Range
    """
    The range at which the message applies.
    """
    severity: Optional[DiagnosticSeverity] = None
    """
    The diagnostic's severity.
    """
    code: Optional[Union[int, str]] = None
    """
    The diagnostic's code.
    """
    code_description: Optional[CodeDescription] = None
    """
    An URI to open with more information about the diagnostic error.
    """
    source: Optional[str] = None
    """
    A human-readable string describing the source of this diagnostic
    """
    message: str
    """
    The diagnostic's message.
    """
    tags: Optional[List[DiagnosticTag]] = None
    """
    Additional metadata about the diagnostic.
    """
    related_information: Optional[List[DiagnosticRelatedInformation]] = None
    """
    An array of related diagnostic information,
    e.g. when symbol-names within a scope collide all definitions can be marked via this property.
    """
    data: Optional[Any] = None
    """
    A data entry field that is preserved between
    a `textDocument/publishDiagnostics` notification and `textDocument/codeAction` request.
    """

    def __members(self) -> Tuple:
        return (
            self.range,
            self.severity,
            self.code,
            self.code_description,
            self.source,
            self.message,
            frozenset(self.tags) if self.tags is not None else None,
            frozenset(self.related_information)
            if self.related_information is not None
            else None,
            self.data,
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__members() == other.__members()
        return NotImplemented

    def __hash__(self):
        return hash(self.__members())


class Command(LspModel):
    title: str
    """
    Title of the command, like `save`.
    """
    command: str
    """
    The identifier of the actual command handler.
    """
    arguments: Optional[List[Any]]
    """
    Arguments that the command handler should be invoked with.
    """


class TextEdit(LspModel):
    range: Range
    """
    The range of the text document to be manipulated.
    To insert text into a document create a range where start === end.
    """
    new_text: str
    """
    The text to be inserted. For delete operations use an empty string.
    """


class ChangeAnnotation(LspModel):
    label: str
    """
    A human-readable string describing the actual change.
    """
    needs_confirmation: Optional[bool]
    """
    A flag which indicates that user confirmation is needed before applying the change.
    """
    description: Optional[str]
    """
    A human-readable string which is rendered less prominent in the user interface.
    """


class AnnotatedTextEdit(TextEdit):
    annotation_id: ChangeAnnotationIdentifier
    """
    The actual annotation identifier.
    """


class TextDocumentIdentifier(LspModel):
    uri: DocumentUri
    """
    The text document's URI.
    """


class VersionedTextDocumentIdentifier(TextDocumentIdentifier):
    version: int
    """
    The version number of this document.
    The version number of a document will increase after each change,
    including undo/redo. The number doesn't need to be consecutive.
    """


class OptionalVersionedTextDocumentIdentifier(TextDocumentIdentifier):
    version: Optional[Union[int, None]]
    """
    The version number of this document. If an optional versioned text document
    identifier is sent from the server to the client and the file is not
    open in the editor (the server has not received an open notification
    before) the server can send `null` to indicate that the version is
    known and the content on disk is the master (as specified with document
    content ownership).
    The version number of a document will increase after each change,
    including undo/redo. The number doesn't need to be consecutive.
    """


class TextDocumentEdit(LspModel):
    text_document: OptionalVersionedTextDocumentIdentifier
    """
    The text document to change.
    """
    edits: List[Union[TextEdit, AnnotatedTextEdit]]
    """
    The edits to be applied.
    """


class CreateFileOptions(LspModel):
    """
    Options to create a file.
    """

    overwrite: Optional[bool]
    """
    Overwrite existing file. Overwrite wins over `ignoreIfExists`
    """
    ignore_if_exists: Optional[bool]
    """
    Ignore if exists.
    """


class CreateFile(LspModel):
    """
    Create file operation.
    """

    kind: str = "create"
    """
    A create.
    """
    uri: DocumentUri
    """
    The resource to create.
    """
    options: Optional[CreateFileOptions]
    """
    Additional options.
    """
    annotation_id: Optional[ChangeAnnotationIdentifier]
    """
    An optional annotation identifier describing the operation.
    """


class RenameFileOptions(LspModel):
    """
    Rename file options.
    """

    overwrite: Optional[bool]
    """
    Overwrite target if existing. Overwrite wins over `ignoreIfExists`.
    """
    ignore_if_exists: Optional[bool]
    """
    Ignores if target exists.
    """


class RenameFile(LspModel):
    """
    Rename file operation.
    """

    kind: str = "rename"
    """
    A rename.
    """
    old_uri: DocumentUri
    """
    The old (existing) location.
    """
    new_uri: DocumentUri
    """
    The new location.
    """
    options: Optional[RenameFileOptions]
    """
    Rename options.
    """
    annotation_id: Optional[ChangeAnnotationIdentifier]
    """
    An optional annotation identifier describing the operation.
    """


class DeleteFileOptions(LspModel):
    """
    Delete file options.
    """

    recursive: Optional[bool]
    """
    Delete the content recursively if a folder denoted.
    """
    ignore_if_not_exists: Optional[bool]
    """
    Ignore the operation if the file does not exists.
    """


class DeleteFile(LspModel):
    """
    Delete file operation
    """

    kind: str = "delete"
    """
    A delete.
    """
    uri: DocumentUri
    """
    The file to delete
    """
    options: Optional[RenameFileOptions]
    """
    Delete options.
    """
    annotation_id: Optional[ChangeAnnotationIdentifier]
    """
    An optional annotation identifier describing the operation.
    """


class WorkspaceEdit(LspModel):
    changes: Optional[Dict[DocumentUri, List[TextEdit]]] = None
    """
    Holds changes to existing resource.
    """
    document_changes: Optional[
        List[Union[TextDocumentEdit, CreateFile, RenameFile, DeleteFile]]
    ] = None
    """
    Depending on the client capability
    `workspace.workspaceEdit.resourceOperations` document changes are either
    an array of `TextDocumentEdit`s to express changes to n different text
    documents where each text document edit applies to a specific version of
    a text document. Or it can contain above `TextDocumentEdit`s mixed with
    create, rename and delete file / folder operations.

    Whether a client supports versioned document edits is expressed via
    `workspace.workspaceEdit.documentChanges` client capability.

    If a client neither supports `documentChanges` nor
    `workspace.workspaceEdit.resourceOperations` then only plain `TextEdit`s
    using the `changes` property are supported.
    """
    change_annotations: Optional[Dict[str, ChangeAnnotation]] = None
    """
    A map of change annotations that can be referenced in
    `AnnotatedTextEdit`s or create, rename, delete file / folder
    operations.

    Whether clients honor this property depends on the client capability
    `workspace.changeAnnotationSupport`.

    @since 3.16.0
    """


class ResourceOperationKind(StrEnum):
    CREATE = "create"
    RENAME = "rename"
    DELETE = "delete"


class FailureHandlingKind(StrEnum):
    ABORT = "abort"
    TRANSACTIONAL = "transactional"
    TEXT_ONLY_TRANSACTIONAL = "textOnlyTransactional"
    UNDO = "undo"


class ChangeAnnotationsSupport(LspModel):
    groups_on_label: Optional[bool]


class WorkspaceEditClientCapabilities(LspModel):
    document_changes: Optional[bool]
    """
    The client supports versioned document changes in `WorkspaceEdit`s
    """
    resource_operations: Optional[List[ResourceOperationKind]]
    """
    The resource operations the client supports.
    Clients should at least support 'create', 'rename' and 'delete' files and folders.
    """
    failure_handling: Optional[FailureHandlingKind]
    """
    The failure handling strategy of a client if applying the workspace edit fails.
    """
    normalizes_line_endings: Optional[bool]
    """
    Whether the client normalizes line endings to the client specific setting.
    If set to `true` the client will normalize line ending characters
    in a workspace edit to the client specific new line character(s).
    """
    # ?
    change_annotation_support: Optional[ChangeAnnotationsSupport]
    """
    Whether the client in general supports change annotations on text edits,
    create file, rename file and delete file changes.
    """


class TextDocumentItem(LspModel):
    uri: DocumentUri
    """
    The text document's URI.
    """
    language_id: str
    """
    The text document's language identifier.
    """
    version: int
    """
    The version number of this document (it will increase after each change, including undo/redo).
    """
    text: str
    """
    The content of the opened text document.
    """


class TextDocumentPositionParams(LspModel):
    text_document: TextDocumentIdentifier
    """
    The text document.
    """
    position: Position
    """
    The position inside the text document.z
    """


class DocumentFilter(LspModel):
    language: Optional[str] = None
    """
    A language ID, like 'typescript'.
    """
    scheme: Optional[str] = None
    """
     A Uri [scheme](#Uri.scheme), like `file` or `untitled`.
    """
    pattern: Optional[str] = None
    """
    Glob patterns can have the following syntax:
    * `*` to match one or more characters in a path segment
    * `?` to match on one character in a path segment
    * `**` to match any number of path segments, including none
    * `{}` to group sub patterns into an OR expression. (e.g. `**​/*.{ts,js}`
        matches all TypeScript and JavaScript files)
    * `[]` to declare a range of characters to match in a path segment
        (e.g., `example.[0-9]` to match on `example.0`, `example.1`, …)
    * `[!...]` to negate a range of characters to match in a path segment
        (e.g., `example.[!0-9]` to match on `example.a`, `example.b`, but
        not `example.0`)
    """


class StaticRegistrationOptions(LspModel):
    """
    Static registration options to be returned in the initialize request.
    """

    id: Optional[str] = None
    """
    The id used to register the request.
    The id can be used to deregister the request again. See also Registration#id.
    """


class TextDocumentRegistrationOptions(LspModel):
    """
    General text document registration options.
    """

    document_selector: Optional[List[DocumentFilter]] = None
    """
    A document selector to identify the scope of the registration.
    If set to null the document selector provided on the client side will be used.
    """


class MarkupKind(StrEnum):
    PLAIN_TEXT = "plaintext"
    MARKDOWN = "markdown"


class MarkupContent(LspModel):
    kind: MarkupKind
    """
    The type of the Markup.
    """
    value: str
    """
    The content itself.
    """


class MarkupClientCapabilities(LspModel):
    parser: MarkupKind
    """
    The name of the parser.
    """
    version: Optional[str]
    """
    The version of the parser.
    """
    allowed_tags: Optional[List[str]]
    """
    A list of HTML tags that the client allows / supports in Markdown.

    Known markdown parsers used by clients right now are:
    -------------------------------------------------------------
    Parser	            Version	    Documentation
    -------------------------------------------------------------
    marked	            1.1.0	    Marked Documentation
    Python-Markdown	    3.2.2	    Python-Markdown Documentation
    -------------------------------------------------------------
    """


class WorkDoneProgressBegin(LspModel):
    kind: str = "begin"
    title: str
    """
    Mandatory title of the progress operation. Used to briefly inform about
    the kind of operation being performed.

    Examples: "Indexing" or "Linking dependencies".
    """
    cancellable: Optional[bool]
    """
    Controls if a cancel button should show to allow the user to cancel the
    long running operation. Clients that don't support cancellation are
    allowed to ignore the setting.
    """
    message: Optional[str]
    """
    Optional, more detailed associated progress message. Contains
    complementary information to the `title`.

    Examples: "3/25 files", "project/src/module2", "node_modules/some_dep".
    If unset, the previous progress message (if any) is still valid.
    """
    percentage: Optional[int]  # uint8
    """
    Optional progress percentage to display (value 100 is considered 100%).
    If not provided infinite progress is assumed and clients are allowed
    to ignore the `percentage` value in subsequent in report notifications.
    """


class WorkDoneProgressReport(LspModel):
    kind: str = "report"
    cancellable: Optional[bool]
    """
    Controls if a cancel button should show to allow the user to cancel the
    long running operation. Clients that don't support cancellation are
    allowed to ignore the setting.
    """
    message: Optional[str]
    """
    Optional, more detailed associated progress message. Contains
    complementary information to the `title`.
    """
    percentage: Optional[int]  # uint8
    """
    Optional progress percentage to display (value 100 is considered 100%).
    If not provided infinite progress is assumed and clients are allowed
    to ignore the `percentage` value in subsequent in report notifications.
    """


class WorkDoneProgressEnd(LspModel):
    kind: str = "end"
    message: Optional[str]
    """
    Optional, a final message indicating to for example indicate the outcome
    of the operation.
    """


class WorkDoneProgressParams(LspModel):
    work_done_token: Optional[ProgressToken]
    """
    An optional token that a server can use to report work done progress.
    """


class WorkDoneProgressOptions(LspModel):
    work_done_progress: Optional[bool] = None


class PartialResultParams(LspModel):
    partial_result_token: Optional[ProgressToken]
    """
    An optional token that a server can use to report partial results (e.g.streaming)
    to the client.
    """


# ##################### Lifecycle Message #####################


class ClientCapabilitiesWorkspaceFileOperation(LspModel):
    dynamic_registration: Optional[bool]
    did_create: Optional[bool]
    will_create: Optional[bool]
    did_rename: Optional[bool]
    will_rename: Optional[bool]
    did_delete: Optional[bool]
    will_delete: Optional[bool]


class InitializeErrorCodes(IntEnum):
    unknown_protocol_version = 1


class InitializeError(LspModel):
    retry: bool


class InitializedParams(LspModel):
    pass


class MessageType(IntEnum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    LOG = 4


class ShowMessageParams(LspModel):
    type: MessageType
    message: str


class ShowMessageRequestClientCapabilitiesMessageActionItem(LspModel):
    additional_properties_support: Optional[bool]


class ShowMessageRequestClientCapabilities(LspModel):
    message_action_item: Optional[ShowMessageRequestClientCapabilitiesMessageActionItem]


class MessageActionItem(LspModel):
    title: str


class ShowMessageRequestParams(LspModel):
    type: MessageType
    message: str
    actions: Optional[List[MessageActionItem]]


class ShowDocumentClientCapabilities(LspModel):
    support: bool


class ShowDocumentParams(LspModel):
    uri: URI
    external: Optional[bool]
    take_focus: Optional[bool]
    selection: Optional[Range]


class ShowDocumentResult(LspModel):
    success: bool
    """
    A boolean indicating if the show was successful.
    """


class LogMessageParams(LspModel):
    type: MessageType
    """
    The message type
    """
    message: str
    """
    The actual message
    """


class PublishDiagnosticsParams(LspModel):
    uri: DocumentUri
    """
    The URI for which diagnostic information is reported.
    """
    version: Optional[int]
    """
    Optional the version number of the document the diagnostics are published for.

    @since 3.15.0
    """
    diagnostics: List[Diagnostic]
    """
    An array of diagnostic information items.
    """


class WorkDoneProgressCreateParams(LspModel):
    token: ProgressToken
    """
    The token to be used to report progress.
    """


class WorkDoneProgressCancelParams(LspModel):
    token: ProgressToken


class Registration(LspModel):
    id: str
    """
    The id used to register the request. The id can be used to deregister
    the request again.
    """
    method: str
    """
    The method / capability to register for.
    """
    registration_options: Optional[Any]
    """
    Options necessary for the registration.
    """


class RegistrationParams(LspModel):
    registrations: List[Registration]


class Unregistration(LspModel):
    id: str
    """
    The id used to unregister the request or notification. Usually an id
    provided during the register request.
    """
    method: str
    """
    The method / capability to unregister for.
    """


class UnregistrationParams(LspModel):
    unregistrations: List[Unregistration]


class SetTraceParams(LspModel):
    value: TraceValue
    """
    The new value that should be assigned to the trace setting.
    """


class LogTraceParams(LspModel):
    message: str
    """
    The message to be logged.
    """
    verbose: Optional[str]
    """
    Additional information that can be computed if the `trace` configuration
    is set to `'verbose'`
    """


class ProgressParams(LspModel):
    token: ProgressToken
    """
    The progress token provided by the client or server.
    """
    value: Any
    """
    The progress data.
    """


class SymbolKind(IntEnum):
    FILE = 1
    MODULE = 2
    NAMESPACE = 3
    PACKAGE = 4
    CLASS = 5
    METHOD = 6
    PROPERTY = 7
    FIELD = 8
    CONSTRUCTOR = 9
    ENUM = 10
    INTERFACE = 11
    FUNCTION = 12
    VARIABLE = 13
    CONSTANT = 14
    STRING = 15
    NUMBER = 16
    BOOLEAN = 17
    ARRAY = 18
    OBJECT = 19
    KEY = 20
    NULL = 21
    ENUMMEMBER = 22
    STRUCT = 23
    EVENT = 24
    OPERATOR = 25
    TYPEPARAMETER = 26


class SymbolTag(IntEnum):
    DEPRECATED = 1


class DocumentSymbol(LspModel):
    name: str
    detail: Optional[str] = None
    kind: SymbolKind
    tags: Optional[List[SymbolTag]] = None
    deprecated: Optional[bool] = None
    range: Range
    selection_range: Range
    children: Optional[List["DocumentSymbol"]] = None


class SymbolInformation(LspModel):
    name: str
    kind: SymbolKind
    tags: Optional[List[SymbolTag]]
    deprecated: Optional[bool]
    location: Location
    container_name: Optional[str]


class WorkspaceSymbolClientCapabilitiesSymbolKind(LspModel):
    value_set: Optional[List[SymbolKind]]


class WorkspaceSymbolClientCapabilitiesTagSupport(LspModel):
    value_set: List[SymbolTag]


class WorkspaceSymbolClientCapabilitiesResolveSupport(LspModel):
    properties: List[str]


class WorkspaceSymbolClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    symbol_kind: Optional[WorkspaceSymbolClientCapabilitiesSymbolKind]
    tag_support: Optional[WorkspaceSymbolClientCapabilitiesTagSupport]
    resolve_support: Optional[WorkspaceSymbolClientCapabilitiesResolveSupport]


class WorkspaceSymbolOptions(WorkDoneProgressOptions):
    resolve_provider: Optional[bool]


class WorkspaceSymbolRegistrationOptions(WorkspaceSymbolOptions):
    pass


class WorkspaceSymbolParams(WorkDoneProgressParams, PartialResultParams):
    query: str


class WorkspaceSymbol(LspModel):
    name: str
    kind: SymbolKind
    tags: Optional[List[SymbolTag]]
    location: Union[Location, DocumentUri]
    container_name: Optional[str]


class ConfigurationItem(LspModel):
    scope_uri: Optional[DocumentUri] = None
    section: Optional[str] = None


class ConfigurationParams(LspModel):
    items: List[ConfigurationItem]


class DidChangeConfigurationClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]


class DidChangeConfigurationParams(LspModel):
    settings: Any


class WorkspaceFolder(LspModel):
    uri: DocumentUri
    name: str


class WorkspaceFoldersChangeEvent(LspModel):
    added: List[WorkspaceFolder]
    removed: List[WorkspaceFolder]


class DidChangeWorkspaceFoldersParams(LspModel):
    event: WorkspaceFoldersChangeEvent


class FileCreate(LspModel):
    uri: str


class CreateFilesParams(LspModel):
    files: List[FileCreate]


class FileRename(LspModel):
    old_uri: str
    new_uri: str


class RenameFilesParams(LspModel):
    files: List[FileRename]


class FileDelete(LspModel):
    uri: str


class DeleteFilesParams(LspModel):
    files: List[FileDelete]


class DidChangeWatchedFilesClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]


class FileSystemWatcher(LspModel):
    glob_pattern: str
    kind: Optional[int]  # uint


class DidChangeWatchedFilesRegistrationOptions(LspModel):
    watchers: List[FileSystemWatcher]


class WatchKind(IntEnum):
    CREATE = 1
    CHANGE = 2
    DELETE = 4


class FileEvent(LspModel):
    uri: DocumentUri
    type: int  # uint


class DidChangeWatchedParams(LspModel):
    changes: List[FileEvent]


class FileChangeType(IntEnum):
    CREATED = 1
    CHANGED = 1
    DELETED = 3


class ExecuteCommandClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]


class ExecuteCommandOptions(WorkDoneProgressOptions):
    commands: List[str]


class ExecuteCommandRegistrationOptions(ExecuteCommandOptions):
    pass


class ExecuteCommandParams(WorkDoneProgressParams):
    command: str
    arguments: Optional[List[Any]]


class ApplyWorkspaceEditParams(LspModel):
    label: Optional[str]
    edit: WorkspaceEdit


class ApplyWorkspaceEditResult(LspModel):
    applied: bool
    failure_reason: Optional[str]
    failed_change: Optional[int]  # uint


class CodeLensWorkspaceClientCapabilities(LspModel):
    refresh_support: Optional[bool]


class SemanticTokensClientCapabilitiesRequestsFull(LspModel):
    delta: Optional[bool]


class SemanticTokensClientCapabilitiesRequests(LspModel):
    range: Optional[bool]
    full: Optional[Union[bool, SemanticTokensClientCapabilitiesRequestsFull]]


class TokenFormat(LspModel):
    pass


class SemanticTokensClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    requests: Optional[SemanticTokensClientCapabilitiesRequests]
    token_types: List[str] = []
    token_modifiers: List[str] = []
    formats: List[TokenFormat] = []
    overlapping_token_support: Optional[bool]
    multiline_token_support: Optional[bool]
    server_cancel_support: Optional[bool]
    augments_syntax_tokens: Optional[bool]


class SemanticTokensWorkspaceClientCapabilities(LspModel):
    refresh_support: Optional[bool]


class TextDocumentSyncClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    will_save: Optional[bool]
    will_save_wait_until: Optional[bool]
    did_save: Optional[bool]


class ClientCapabilitiesWorkspace(LspModel):
    apply_edit: Optional[bool]
    workspace_edit: Optional[WorkspaceEditClientCapabilities]
    did_change_configuration: Optional[DidChangeConfigurationClientCapabilities]
    did_change_watched_files: Optional[DidChangeWatchedFilesClientCapabilities]
    symbol: Optional[WorkspaceSymbolClientCapabilities]
    execute_command: Optional[ExecuteCommandClientCapabilities]
    workspace_folders: Optional[bool]
    configuration: Optional[bool]
    semantic_tokens: Optional[SemanticTokensWorkspaceClientCapabilities]
    code_lens: Optional[CodeLensWorkspaceClientCapabilities]
    file_operations: Optional[ClientCapabilitiesWorkspaceFileOperation]


class TextDocumentClientCapabilities(LspModel):
    synchronization: Optional[TextDocumentSyncClientCapabilities]
    """
    completion: Optional[CompletionClientCapabilities]
    hover: Optional[HoverClientCapabilities]
    signature_help: Optional[SignatureHelpClientCapabilities]
    declaration: Optional[DeclarationClientCapabilities]
    definition: Optional[DefinitionClientCapabilities]
    type_definition: Optional[TypeDefinitionClientCapabilities]
    implementation: Optional[ImplementationClientCapabilities]
    references: Optional[ReferenceClientCapabilities]
    document_highlight: Optional[DocumentHighlightClientCapabilities]
    document_symbol: Optional[DocumentSymbolClientCapabilities]
    code_action: Optional[CodeActionClientCapabilities]
    document_link: Optional[DocumentLinkClientCapabilities]
    color_provider: Optional[DocumentColorClientCapabilities]
    formatting: Optional[DocumentFormattingClientCapabilities]
    range_formatting: Optional[DocumentRangeFormattingClientCapabilities]
    on_type_formatting: Optional[DocumentOnTypeFormattingClientCapabilities]
    rename: Optional[RenameClientCapabilities]
    publish_diagnostic: Optional[PublishDiagnosticsClientCapabilities]
    folding_range: Optional[FoldingRangeClientCapabilities]
    selection_range: Optional[SelectionRangeClientCapabilities]
    linked_editing_range: Optional[LinkedEditingRangeClientCapabilities]
    call_hierarchy: Optional[CallHierarchyClientCapabilities]
    semantic_tokens: Optional[SemanticTokensClientCapabilities]
    moniker: Optional[MonikerClientCapabilities]
    type_hierarchy: Optional[TypeHierarchyClientCapabilities]
    inline_valie: Optional[InlineValueClientCapabilities]
    inlay_hint: Optional[InlayHintClientCapabilities]
    """


class ClientCapabilitiesGeneralStaleRequestSupport(LspModel):
    cancel: bool
    retry_on_content_modified: List[str]


class PositionEncodingKind(StrEnum):
    UTF8 = "utf-8"
    UTF16 = "utf-16"
    UTF32 = "utf-32"


class ClientCapabilitiesGeneral(LspModel):
    stale_request_support: Optional[ClientCapabilitiesGeneralStaleRequestSupport]
    regular_expressions: Optional[RegularExpressionsClientCapabilities]
    position_encodings: Optional[List[PositionEncodingKind]]
    """
    markdown: Optional[MarkdownClientCapabilities]
    """


class ClientCapabilitiesWindow(LspModel):
    work_done_progress: Optional[bool]
    show_message: Optional[ShowMessageRequestClientCapabilities]
    show_document: Optional[ShowDocumentClientCapabilities]


class NotebookDocumentSyncClientCapabilities(LspModel):
    dynamic_registration: Optional[bool]
    execution_summary_support: Optional[bool]


class NotebookDocumentClientCapabilities(LspModel):
    synchronization: NotebookDocumentSyncClientCapabilities


class ClientCapabilities(LspModel):
    workspace: Optional[ClientCapabilitiesWorkspace]
    text_document: Optional[TextDocumentClientCapabilities]
    notebook_document: Optional[NotebookDocumentClientCapabilities]
    window: Optional[ClientCapabilitiesWindow]
    general: Optional[ClientCapabilitiesGeneral]
    experimental: Optional[Any]


class InitializeParamsClientInfo(LspModel):
    name: str
    version: Optional[str]


class InitializeParams(LspModel):
    process_id: Optional[int]
    client_info: Optional[InitializeParamsClientInfo]
    locale: Optional[str]
    root_path: Optional[str]
    root_uri: Optional[DocumentUri]
    initialization_options: Optional[Any]
    capabilities: ClientCapabilities
    trace: Optional[TraceValue]
    workspace_folders: Optional[List[WorkspaceFolder]]



from . import enums, types
from .abc import IrAbc, SolidityAbc
from .declarations.abc import DeclarationAbc
from .declarations.contract_definition import ContractDefinition
from .declarations.enum_definition import EnumDefinition
from .declarations.enum_value import EnumValue
from .declarations.error_definition import ErrorDefinition
from .declarations.event_definition import EventDefinition
from .declarations.function_definition import FunctionDefinition
from .declarations.modifier_definition import ModifierDefinition
from .declarations.struct_definition import StructDefinition
from .declarations.user_defined_value_type_definition import (
    UserDefinedValueTypeDefinition,
)
from .declarations.variable_declaration import VariableDeclaration
from .expressions.abc import ExpressionAbc
from .expressions.assignment import AssignedVariablePath, Assignment
from .expressions.binary_operation import BinaryOperation
from .expressions.conditional import Conditional
from .expressions.elementary_type_name_expression import ElementaryTypeNameExpression
from .expressions.function_call import FunctionCall
from .expressions.function_call_options import FunctionCallOptions
from .expressions.identifier import Identifier
from .expressions.index_access import IndexAccess
from .expressions.index_range_access import IndexRangeAccess
from .expressions.literal import Literal
from .expressions.member_access import MemberAccess
from .expressions.new_expression import NewExpression
from .expressions.tuple_expression import TupleExpression
from .expressions.unary_operation import UnaryOperation
from .meta.identifier_path import IdentifierPath, IdentifierPathPart
from .meta.import_directive import ImportDirective
from .meta.inheritance_specifier import InheritanceSpecifier
from .meta.modifier_invocation import ModifierInvocation
from .meta.override_specifier import OverrideSpecifier
from .meta.parameter_list import ParameterList
from .meta.pragma_directive import PragmaDirective
from .meta.source_unit import SourceUnit
from .meta.structured_documentation import StructuredDocumentation
from .meta.try_catch_clause import TryCatchClause
from .meta.using_for_directive import UsingForDirective
from .statements.abc import StatementAbc
from .statements.block import Block
from .statements.break_statement import Break
from .statements.continue_statement import Continue
from .statements.do_while_statement import DoWhileStatement
from .statements.emit_statement import EmitStatement
from .statements.expression_statement import ExpressionStatement
from .statements.for_statement import ForStatement
from .statements.if_statement import IfStatement
from .statements.inline_assembly import ExternalReference, InlineAssembly
from .statements.placeholder_statement import PlaceholderStatement
from .statements.return_statement import Return
from .statements.revert_statement import RevertStatement
from .statements.try_statement import TryStatement
from .statements.unchecked_block import UncheckedBlock
from .statements.variable_declaration_statement import VariableDeclarationStatement
from .statements.while_statement import WhileStatement
from .type_names.abc import TypeNameAbc
from .type_names.array_type_name import ArrayTypeName
from .type_names.elementary_type_name import ElementaryTypeName
from .type_names.function_type_name import FunctionTypeName
from .type_names.mapping import Mapping
from .type_names.user_defined_type_name import UserDefinedTypeName
from .yul.abc import YulAbc, YulStatementAbc
from .yul.assignment import YulAssignment
from .yul.block import YulBlock
from .yul.break_statement import YulBreak
from .yul.case_statement import YulCase
from .yul.continue_statement import YulContinue
from .yul.expression_statement import YulExpressionStatement
from .yul.for_loop import YulForLoop
from .yul.function_call import YulFunctionCall
from .yul.function_definition import YulFunctionDefinition
from .yul.identifier import YulIdentifier
from .yul.if_statement import YulIf
from .yul.leave import YulLeave
from .yul.literal import YulLiteral
from .yul.switch import YulSwitch
from .yul.typed_name import YulTypedName
from .yul.variable_declaration import YulVariableDeclaration



from functools import lru_cache, reduce
from operator import or_
from typing import Iterator, List, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcFunctionCallOptions
from woke.ir.enums import ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.utils import IrInitTuple


class FunctionCallOptions(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcFunctionCallOptions
    _parent: SolidityAbc  # TODO: make this more specific

    _expression: ExpressionAbc
    _names: List[str]
    _options: List[ExpressionAbc]

    def __init__(
        self,
        init: IrInitTuple,
        function_call_options: SolcFunctionCallOptions,
        parent: SolidityAbc,
    ):
        super().__init__(init, function_call_options, parent)
        self._expression = ExpressionAbc.from_ast(
            init, function_call_options.expression, self
        )
        self._names = list(function_call_options.names)
        self._options = [
            ExpressionAbc.from_ast(init, option, self)
            for option in function_call_options.options
        ]

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._expression
        for option in self._options:
            yield from option

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def expression(self) -> ExpressionAbc:
        return self._expression

    @property
    def names(self) -> Tuple[str, ...]:
        return tuple(self._names)

    @property
    def options(self) -> Tuple[ExpressionAbc, ...]:
        return tuple(self._options)

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        ret = self.expression.modifies_state | reduce(
            or_,
            (option.modifies_state for option in self.options),
            set(),
        )
        if "value" in self.names:
            ret |= {(self, ModifiesStateFlag.SENDS_ETHER)}
        return ret



from __future__ import annotations

from collections import deque
from typing import Deque, Dict, List, Optional, Set, Union

import woke.ir as ir
import woke.ir.types as types
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


class AxelarProxyContractIdDetector(Detector):
    proxies: Dict[bytes, Set[ir.FunctionDefinition]]
    upgradeables: Dict[bytes, Set[ir.FunctionDefinition]]

    def __init__(self):
        self.proxies = {}
        self.upgradeables = {}

    @detector.command("axelar-proxy-contract-id")
    def cli(self):
        """
        Detects incorrect use of the `contractId` function in Axelar proxy and upgradeable contracts.
        """
        pass

    def detect(self) -> List[DetectorResult]:
        proxy_contracts: Dict[bytes, Set[ir.ContractDefinition]] = {}
        for selector, fns in self.proxies.items():
            proxy_contracts[selector] = set()
            visited: Set[
                ir.ContractDefinition
            ] = set(  # pyright: ignore reportGeneralTypeIssues
                fn.parent for fn in fns
            )
            queue: Deque[
                ir.ContractDefinition
            ] = deque(  # pyright: ignore reportGeneralTypeIssues
                fn.parent for fn in fns
            )

            while len(queue) > 0:
                contract = queue.popleft()
                if (
                    contract.kind == ir.enums.ContractKind.CONTRACT
                    and not contract.abstract
                ):
                    proxy_contracts[selector].add(contract)
                for child in contract.child_contracts:
                    if child not in visited and (
                        len(child.functions) == 0
                        or not any(fn.name == "contractId" for fn in child.functions)
                    ):
                        visited.add(child)
                        queue.append(child)

        upgradeable_contracts: Dict[bytes, Set[ir.ContractDefinition]] = {}
        for selector, fns in self.upgradeables.items():
            upgradeable_contracts[selector] = set()
            visited: Set[
                ir.ContractDefinition
            ] = set(  # pyright: ignore reportGeneralTypeIssues
                fn.parent for fn in fns
            )
            queue: Deque[
                ir.ContractDefinition
            ] = deque(  # pyright: ignore reportGeneralTypeIssues
                fn.parent for fn in fns
            )

            while len(queue) > 0:
                contract = queue.popleft()
                if (
                    contract.kind == ir.enums.ContractKind.CONTRACT
                    and not contract.abstract
                ):
                    upgradeable_contracts[selector].add(contract)
                for child in contract.child_contracts:
                    if child not in visited and (
                        len(child.functions) == 0
                        or not any(fn.name == "contractId" for fn in child.functions)
                    ):
                        visited.add(child)
                        queue.append(child)

        ret = []

        for proxies in sorted(proxy_contracts.values()):
            if len(proxies) > 1:
                sorted_proxies = sorted(proxies, key=lambda c: c.name)
                ret.append(
                    Detection(
                        sorted_proxies[0],
                        "Proxy contract ID is shared between multiple contracts",
                        tuple(
                            Detection(
                                other, "Other contract", lsp_range=other.name_location
                            )
                            for other in sorted_proxies[1:]
                        ),
                        lsp_range=sorted_proxies[0].name_location,
                    )
                )

        for only_proxy in sorted(proxy_contracts.keys() - upgradeable_contracts.keys()):
            for proxy in sorted(proxy_contracts[only_proxy], key=lambda c: c.name):
                ret.append(
                    Detection(
                        proxy,
                        "Found a proxy contract without an upgradeable contract with the same contract ID",
                        lsp_range=proxy.name_location,
                    )
                )
        for only_upgradeable in sorted(
            upgradeable_contracts.keys() - proxy_contracts.keys()
        ):
            for upgradeable in sorted(
                upgradeable_contracts[only_upgradeable], key=lambda c: c.name
            ):
                ret.append(
                    Detection(
                        upgradeable,
                        "Found an upgradeable contract without a proxy contract with the same contract ID",
                        lsp_range=upgradeable.name_location,
                    )
                )

        whitelisted_functions = {
            bytes.fromhex("5c60da1b"): "implementation",
            bytes.fromhex("9ded06df"): "setup",
        }

        for common_contract_id in sorted(
            proxy_contracts.keys() & upgradeable_contracts.keys()
        ):
            for proxy in sorted(
                proxy_contracts[common_contract_id], key=lambda c: c.name
            ):
                for upgradeable in sorted(
                    upgradeable_contracts[common_contract_id], key=lambda c: c.name
                ):
                    proxy_functions: Dict[bytes, ir.FunctionDefinition] = {}
                    upgradeable_functions: Dict[bytes, ir.FunctionDefinition] = {}

                    for contract in proxy.linearized_base_contracts:
                        for func in contract.functions:
                            if (
                                func.function_selector is not None
                                and func.function_selector not in proxy_functions
                            ):
                                proxy_functions[func.function_selector] = func

                    for contract in upgradeable.linearized_base_contracts:
                        for func in contract.functions:
                            if (
                                func.function_selector is not None
                                and func.function_selector not in upgradeable_functions
                            ):
                                upgradeable_functions[func.function_selector] = func

                    for common_selector in sorted(
                        proxy_functions.keys() & upgradeable_functions.keys()
                    ):
                        if common_selector in whitelisted_functions:
                            func_name = whitelisted_functions[common_selector]
                            if (
                                proxy_functions[common_selector].name == func_name
                                and upgradeable_functions[common_selector].name
                                == func_name
                            ):
                                continue
                        ret.append(
                            Detection(
                                proxy,
                                "Found a proxy contract and an upgradeable contract with a function with the same function selector",
                                (
                                    Detection(
                                        upgradeable,
                                        "Upgradeable contract",
                                        lsp_range=upgradeable.name_location,
                                    ),
                                    Detection(
                                        proxy_functions[common_selector],
                                        f"Proxy contract function (selector: {common_selector.hex()})",
                                        lsp_range=proxy_functions[
                                            common_selector
                                        ].name_location,
                                    ),
                                    Detection(
                                        upgradeable_functions[common_selector],
                                        f"Upgradeable contract function (selector: {common_selector.hex()})",
                                        lsp_range=upgradeable_functions[
                                            common_selector
                                        ].name_location,
                                    ),
                                ),
                                lsp_range=proxy.name_location,
                            )
                        )

        return [
            DetectorResult(
                d, impact=DetectionImpact.WARNING, confidence=DetectionConfidence.HIGH
            )
            for d in ret
        ]

    def _eval_contract_id(self, expr: ir.ExpressionAbc) -> Optional[Union[bytes, int]]:
        from Crypto.Hash import keccak

        if isinstance(expr, ir.FunctionCall):
            if expr.kind == ir.enums.FunctionCallKind.TYPE_CONVERSION:
                arg = self._eval_contract_id(expr.arguments[0])
                if isinstance(expr.type, types.IntAbc):
                    if isinstance(arg, int):
                        return arg
                    elif isinstance(arg, bytes):
                        return int.from_bytes(arg, "big", signed=False)
                    else:
                        return None
                elif isinstance(expr.type, (types.Bytes, types.FixedBytes)):
                    if isinstance(arg, bytes):
                        return arg
                    elif isinstance(arg, int):
                        return arg.to_bytes(32, "big", signed=False)
                    else:
                        return None
            elif expr.function_called == ir.enums.GlobalSymbolsEnum.KECCAK256:
                arg = expr.arguments[0]
                if not isinstance(arg, ir.Literal) or not isinstance(
                    arg.type, types.StringLiteral
                ):
                    return None
                contract_id = arg.value
                if contract_id is None:
                    return None

                return keccak.new(
                    data=contract_id.encode("utf-8"), digest_bits=256
                ).digest()
        elif isinstance(expr, ir.Literal):
            if expr.kind == ir.enums.LiteralKind.HEX_STRING:
                return bytes.fromhex(expr.hex_value)
            elif expr.kind == ir.enums.LiteralKind.NUMBER and expr.value is not None:
                if expr.value.startswith("0x"):
                    str_val = expr.value[2:].replace("_", "")
                    try:
                        return int(str_val, 16)
                    except ValueError:
                        return None
                else:
                    try:
                        return int(expr.value.replace("_", ""))
                    except ValueError:
                        return None
            else:
                return None
        elif isinstance(expr, ir.BinaryOperation):
            left = self._eval_contract_id(expr.left_expression)
            right = self._eval_contract_id(expr.right_expression)
            if expr.operator == ir.enums.BinaryOpOperator.PLUS:
                if isinstance(left, int) and isinstance(right, int):
                    return left + right
                else:
                    return None
            elif expr.operator == ir.enums.BinaryOpOperator.MINUS:
                if isinstance(left, int) and isinstance(right, int):
                    return left - right
                else:
                    return None
            else:
                return None
        else:
            return None

    def visit_contract_definition(self, node: ir.ContractDefinition):
        contract_id_function: Optional[ir.FunctionDefinition] = None
        for func in node.functions:
            if (
                func.name == "contractId"
                and len(func.parameters.parameters) == 0
                and len(func.return_parameters.parameters) == 1
                and isinstance(
                    func.return_parameters.parameters[0].type, types.FixedBytes
                )
                and func.return_parameters.parameters[0].type.bytes_count == 32
                and func.implemented
            ):
                contract_id_function = func
                break
        if contract_id_function is None:
            return

        returns = []
        assert contract_id_function.body is not None

        for statement in contract_id_function.body.statements_iter():
            if isinstance(statement, ir.Return):
                returns.append(statement)

        if len(returns) > 1:
            return

        if len(returns) == 1 and returns[0].expression is not None:
            expr = returns[0].expression
        else:
            ret_var = contract_id_function.return_parameters.parameters[0]
            assignments = []

            for ref in ret_var.references:
                if isinstance(ref, ir.ExternalReference):
                    continue
                elif isinstance(ref, ir.IdentifierPathPart):
                    ref = ref.underlying_node
                assignment = ref
                while assignment is not None:
                    if isinstance(assignment, ir.Assignment):
                        break
                    assignment = assignment.parent
                if assignment is None:
                    continue

                if len(assignment.assigned_variables) != 1:
                    continue
                assigned_var = assignment.assigned_variables[0]
                if assigned_var is None or len(assigned_var) != 1:
                    continue
                path = next(iter(assigned_var))
                if ret_var in path:
                    assignments.append(assignment)

            if len(assignments) != 1:
                return
            expr = assignments[0].right_expression

        contract_id = self._eval_contract_id(expr)
        if contract_id is None:
            return
        elif isinstance(contract_id, bytes) and len(contract_id) != 32:
            contract_id = bytes(32 - len(contract_id)) + contract_id
        elif isinstance(contract_id, int):
            contract_id = contract_id.to_bytes(32, "big", signed=False)

        if contract_id_function.visibility == ir.enums.Visibility.INTERNAL:
            if contract_id not in self.proxies:
                self.proxies[contract_id] = set()
            self.proxies[contract_id].add(contract_id_function)
        else:
            if contract_id not in self.upgradeables:
                self.upgradeables[contract_id] = set()
            self.upgradeables[contract_id].add(contract_id_function)



import logging
from itertools import chain
from typing import Any, List, Optional

from woke.compiler.exceptions import CompilationResolveError
from woke.compiler.source_path_resolver import SourcePathResolver
from woke.compiler.source_unit_name_resolver import SourceUnitNameResolver
from woke.lsp.common_structures import (
    DocumentUri,
    PartialResultParams,
    Position,
    Range,
    TextDocumentIdentifier,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_data_model import LspModel
from woke.lsp.utils.uri import path_to_uri, uri_to_path

logger = logging.getLogger(__name__)


class DocumentLinkOptions(WorkDoneProgressOptions):
    resolve_provider: Optional[bool]
    """
    Document links have a resolve provider as well.
    """


class DocumentLinkParams(WorkDoneProgressParams, PartialResultParams):
    text_document: TextDocumentIdentifier
    """
    The document to provide document links for.
    """


class DocumentLink(LspModel):
    range: Range
    """
    The range this link applies to.
    """
    target: Optional[DocumentUri]
    """
    The uri this link points to. If missing a resolve request is sent later.
    """
    tooltip: Optional[str]
    """
    The tooltip text when you hover over this link.
    
    If a tooltip is provided, it will be displayed in a string that includes
    instructions on how to trigger the link, such as `{0} (ctrl + click)`.
    The specific instructions vary depending on OS, user settings, and
    localization.
    
    @since 3.15.0
    """
    data: Optional[Any]
    """
    A data entry field that is preserved on a document link between a
    DocumentLinkRequest and a DocumentLinkResolveRequest.
    """


async def document_link(
    context: LspContext, params: DocumentLinkParams
) -> Optional[List[DocumentLink]]:
    logger.debug(f"Requested document links for file {params.text_document.uri}")

    path = uri_to_path(params.text_document.uri).resolve()

    document_links = []
    source_unit_name_resolver = SourceUnitNameResolver(context.config)
    source_path_resolver = SourcePathResolver(context.config)

    this_source_unit_name = None
    for include_path in chain(
        context.config.compiler.solc.include_paths, [context.config.project_root_path]
    ):
        try:
            rel_path = str(path.relative_to(include_path).as_posix())
            if this_source_unit_name is None or len(this_source_unit_name) > len(
                rel_path
            ):
                this_source_unit_name = rel_path
        except ValueError:
            continue

    if this_source_unit_name is None:
        return None

    try:
        root = context.parser[path].root_node
    except KeyError:
        return None

    for child in root.children:
        if child.type == "import_directive":
            source_node = child.child_by_field_name("source")

            if source_node is not None:
                import_str = source_node.text.decode("utf-16-le")[1:-1]  # remove quotes
                unit_name = source_unit_name_resolver.resolve_import(
                    this_source_unit_name, import_str
                )

                try:
                    include_path = source_path_resolver.resolve(
                        unit_name, this_source_unit_name, context.parser.files
                    )
                    document_links.append(
                        DocumentLink(
                            range=Range(
                                start=Position(
                                    line=source_node.start_point[0],
                                    character=source_node.start_point[1] // 2,
                                ),
                                end=Position(
                                    line=source_node.end_point[0],
                                    character=source_node.end_point[1] // 2,
                                ),
                            ),
                            target=DocumentUri(path_to_uri(include_path)),
                            tooltip=None,
                            data=None,
                        )
                    )
                except CompilationResolveError:
                    continue

    return document_links



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, List

from woke.core.visitor import Visitor, visit_map

if TYPE_CHECKING:
    from rich.console import Console


class Printer(Visitor, metaclass=ABCMeta):
    console: Console
    paths: List[Path]

    @abstractmethod
    def print(self) -> None:
        ...

    def _run(self) -> None:
        from woke.utils.file_utils import is_relative_to

        for path, source_unit in self.build.source_units.items():
            if len(self.paths) == 0 or any(is_relative_to(path, p) for p in self.paths):
                for node in source_unit:
                    visit_map[node.ast_node.node_type](self, node)

        self.print()



from typing import Iterator, Set, Tuple

from woke.ir.ast import SolcElementaryTypeNameExpression
from woke.ir.enums import ModifiesStateFlag
from woke.ir.utils import IrInitTuple

from ..abc import IrAbc, SolidityAbc
from ..type_names.elementary_type_name import ElementaryTypeName
from .abc import ExpressionAbc


class ElementaryTypeNameExpression(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcElementaryTypeNameExpression
    _parent: SolidityAbc  # TODO: make this more specific

    _type_name: ElementaryTypeName

    def __init__(
        self,
        init: IrInitTuple,
        elementary_type_name_expression: SolcElementaryTypeNameExpression,
        parent: SolidityAbc,
    ):
        super().__init__(init, elementary_type_name_expression, parent)
        self._type_name = ElementaryTypeName(
            init, elementary_type_name_expression.type_name, self
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        yield from self._type_name

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def type_name(self) -> ElementaryTypeName:
        return self._type_name

    @property
    def is_ref_to_state_variable(self) -> bool:
        return False

    @property
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return set()



from .context_managers import change_cwd, recursion_guard
from .decorators import return_on_recursion
from .enums import StrEnum
from .general import get_class_that_defined_method
from .version import get_package_version



from __future__ import annotations

from typing import TYPE_CHECKING

from .abc import YulStatementAbc

if TYPE_CHECKING:
    from .block import YulBlock


class YulBreak(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock

    @property
    def parent(self) -> YulBlock:
        return self._parent



from collections import UserString


class StringReader(UserString):
    __original: str

    def __init__(self, original: str):
        self.__original = original
        super().__init__(original)

    def read(self, prefix: str) -> None:
        if self.startswith(prefix):
            self.data = self.data[len(prefix) :]
        else:
            raise ValueError(
                f"String does not start with '{prefix}'. Original: {self.__original}"
            )

    def insert(self, prefix: str) -> None:
        self.data = prefix + self.data

    @property
    def original(self) -> str:
        return self.__original



import re
from pathlib import Path
from typing import List, Tuple

from Crypto.Hash import BLAKE2b

from woke.core.solidity_version import (
    SolidityVersionExpr,
    SolidityVersionRange,
    SolidityVersionRanges,
)

from .solidity_import import SolidityImportExpr

# TODO Raise error on `pragma` or `import` in contract definition
# Instead of parsing `pragma solidity` and `import` statements inside a contract definition, we should raise an exception.
# example:
# ```
# contract b {
#     pragma solidity ^0.8.0;
# }
# ```
# assignees: michprev


class SoliditySourceParser:
    PRAGMA_SOLIDITY_RE = re.compile(r"pragma\s+solidity\s+(?P<version>[^;]+)\s*;")
    IMPORT_RE = re.compile(r"""import\s*(?P<import>[\s"*{][^;]+)\s*;""")
    MULTILINE_COMMENT_END_RE = re.compile(r"\*/")
    ONELINE_COMMENT_OR_MULTILINE_COMMENT_START_RE = re.compile(
        r"(//.*$|/\*)", re.MULTILINE
    )

    @classmethod
    def __string_closed(cls, line: str) -> bool:
        opening_char = None
        for i in range(len(line)):
            if opening_char is None:
                if line[i] in {'"', "'"}:
                    opening_char = line[i]
            else:
                if line[i] == opening_char:
                    if i > 0 and line[i - 1] == "\\":
                        continue
                    else:
                        opening_char = None
        return opening_char is None

    @classmethod
    def __parse_version_pragma(
        cls, source_code: str, ignore_errors: bool
    ) -> SolidityVersionRanges:
        versions = None
        matches = cls.PRAGMA_SOLIDITY_RE.finditer(source_code)
        for match in matches:
            s = source_code[0 : match.start()].splitlines()
            if len(s) > 0:
                # ignore pragmas in a string
                if not cls.__string_closed(s[-1]):
                    continue

            version_str = match.groupdict()["version"]
            try:
                version_expr = SolidityVersionExpr(version_str)
            except ValueError:
                if ignore_errors:
                    continue
                raise
            if versions is None:
                versions = version_expr.version_ranges
            else:
                # in case of multiple version pragmas in a single file, intersection is performed
                versions &= version_expr.version_ranges

        # any version can be used when no pragma solidity present
        if versions is None:
            versions = SolidityVersionRanges(
                [SolidityVersionRange("0.0.0", True, None, None)]
            )
        return versions

    @classmethod
    def __parse_import(cls, source_code: str, ignore_errors: bool) -> List[str]:
        imports = set()  # avoid listing the same import multiple times
        matches = cls.IMPORT_RE.finditer(source_code)
        for match in matches:
            s = source_code[0 : match.start()].splitlines()
            if len(s) > 0:
                # ignore imports in a string
                if not cls.__string_closed(s[-1]):
                    continue

            import_str = match.groupdict()["import"]
            try:
                import_expr = SolidityImportExpr(import_str)
            except ValueError:
                if ignore_errors:
                    continue
                raise
            imports.add(import_expr.filename)
        return list(imports)

    @classmethod
    def strip_comments(cls, source_code: str, ignore_errors: bool = False) -> str:
        stop = False
        while not stop:
            # try to find a single-line or multi-line comment (whichever comes first)
            matches = cls.ONELINE_COMMENT_OR_MULTILINE_COMMENT_START_RE.finditer(
                source_code
            )
            stop = True

            for match in matches:
                s = source_code[0 : match.start()].splitlines()
                if len(s) > 0:
                    # ignore `//` and `/*` in Solidity strings
                    if not cls.__string_closed(s[-1]):
                        continue

                stop = False

                if source_code[match.start() : match.end()] == "/*":
                    # found a multi-line comment start
                    end_match = cls.MULTILINE_COMMENT_END_RE.search(
                        source_code, match.end()
                    )
                    if end_match is None:
                        if ignore_errors:
                            source_code = source_code[: match.start()]
                            continue
                        raise ValueError(f"Multi-line comment not closed.")
                    source_code = (
                        source_code[0 : match.start()] + source_code[end_match.end() :]
                    )
                else:
                    # found a single-line comment
                    source_code = (
                        source_code[0 : match.start()] + source_code[match.end() :]
                    )

                break
        return source_code

    @classmethod
    def parse(
        cls, path: Path, ignore_errors: bool = False
    ) -> Tuple[SolidityVersionRanges, List[str], bytes, str]:
        """
        Return a tuple of two lists. The first list contains Solidity version ranges that can be used to compile
        the given file. The second list contains filenames / URLs that are imported from the given file.
        """
        raw_content = path.read_bytes()
        content = raw_content.decode("utf-8")

        h = BLAKE2b.new(data=raw_content, digest_bits=256)

        # strip all comments
        stripped_content = cls.strip_comments(content, ignore_errors)

        return (
            cls.__parse_version_pragma(stripped_content, ignore_errors),
            cls.__parse_import(stripped_content, ignore_errors),
            h.digest(),
            content,
        )

    @classmethod
    def parse_source(
        cls,
        source_code: str,
        ignore_errors: bool = False,
    ) -> Tuple[SolidityVersionRanges, List[str], bytes]:
        """
        Return a tuple of two lists. The first list contains Solidity version ranges that can be used to compile
        the given file. The second list contains filenames / URLs that are imported from the given file.
        """
        h = BLAKE2b.new(data=source_code.encode("utf-8"), digest_bits=256)

        # strip all comments
        content = cls.strip_comments(source_code, ignore_errors)

        return (
            cls.__parse_version_pragma(content, ignore_errors),
            cls.__parse_import(content, ignore_errors),
            h.digest(),
        )



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Iterator, Optional, Set, Tuple, Union

from woke.ir.enums import ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc

from .abc import StatementAbc

if TYPE_CHECKING:
    from ..meta.parameter_list import ParameterList

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import AstNodeId, SolcReturn
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class Return(StatementAbc):
    """
    !!! example
        `:::solidity return 1` in the following code:
        ```solidity
        function f() public pure returns(uint) {
            return 1;
        }
        ```
    """

    _ast_node: SolcReturn
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    _function_return_parameters: Optional[AstNodeId]
    _expression: Optional[ExpressionAbc]

    def __init__(self, init: IrInitTuple, return_: SolcReturn, parent: SolidityAbc):
        super().__init__(init, return_, parent)
        self._function_return_parameters = return_.function_return_parameters
        self._expression = (
            ExpressionAbc.from_ast(init, return_.expression, self)
            if return_.expression
            else None
        )

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        if self._expression is not None:
            yield from self._expression

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def function_return_parameters(self) -> Optional[ParameterList]:
        """
        !!! example
            Can be `None` if the function does not return anything.
            ```solidity
            function f(uint x) public {
                if (x > 0) {
                    return;
                }
                doSomething(x);
            }
            ```

        Returns:
            Parameter list describing the return parameters of the function (if any).
        """
        from ..meta.parameter_list import ParameterList

        if self._function_return_parameters is None:
            return None
        node = self._reference_resolver.resolve_node(
            self._function_return_parameters, self._cu_hash
        )
        assert isinstance(node, ParameterList)
        return node

    @property
    def expression(self) -> Optional[ExpressionAbc]:
        """
        Returns:
            Expression returned by the return statement, if any.
        """
        return self._expression

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        if self._expression is None:
            return set()
        return self._expression.modifies_state



import hashlib
import logging
import platform
import shutil
import urllib.error
import urllib.request
from pathlib import Path
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union
from zipfile import ZipFile

import aiohttp
from Crypto.Hash import keccak
from pydantic import BaseModel, Field

from woke.config import UnsupportedPlatformError, WokeConfig
from woke.core.solidity_version import SolidityVersion

from .abc import CompilerVersionManagerAbc
from .exceptions import ChecksumError, UnsupportedVersionError

logger = logging.getLogger(__name__)


class SolcBuildInfo(BaseModel):
    path: str
    version: SolidityVersion
    build: str
    long_version: SolidityVersion = Field(alias="longVersion")
    keccak256: str
    sha256: str
    urls: List[str]


class SolcBuilds(BaseModel):
    builds: List[SolcBuildInfo]
    releases: Dict[SolidityVersion, str]
    latest_release: str = Field(alias="latestRelease")


class SolcVersionManager(CompilerVersionManagerAbc):
    """
    Solc version manager that can install, remove and provide info about `solc` compiler.
    """

    # TODO: Add support for older solc versions in svm module
    #  Currently only builds present in binaries.soliditylang.org repository are supported.
    #  We should also support older solc releases.
    # assignees: michprev

    BINARIES_URL: str = "https://binaries.soliditylang.org"

    __platform: str
    __solc_list_url: str
    __compilers_path: Path
    __solc_list_path: Path
    __solc_builds: Optional[SolcBuilds]

    def __init__(self, woke_config: WokeConfig):
        system = platform.system()
        if system == "Linux":
            self.__platform = "linux-amd64"
        elif system == "Darwin":
            self.__platform = "macosx-amd64"
        elif system == "Windows":
            self.__platform = "windows-amd64"
        else:
            raise UnsupportedPlatformError(f"Platform `{system}` is not supported.")

        self.__solc_list_url = f"{self.BINARIES_URL}/{self.__platform}/list.json"
        self.__compilers_path = woke_config.global_data_path / "compilers"
        self.__solc_list_path = self.__compilers_path / "solc.json"
        self.__solc_builds = None

        self.__compilers_path.mkdir(parents=True, exist_ok=True)

    def installed(self, version: Union[SolidityVersion, str]) -> bool:
        self.__fetch_list_file()
        if self.__solc_builds is None:
            raise RuntimeError(
                f"Unable to fetch or correctly parse '{self.__solc_list_url}'."
            )

        if isinstance(version, str):
            version = SolidityVersion.fromstring(version)

        path = self.get_path(version)
        if not path.is_file():
            return False

        if not self.__verify_checksums(version):
            return False
        return True

    async def install(
        self,
        version: Union[SolidityVersion, str],
        force_reinstall: bool = False,
        http_session: Optional[aiohttp.ClientSession] = None,
        progress: Optional[Callable[[int, int], Awaitable[None]]] = None,
    ) -> None:
        self.__fetch_list_file()
        if self.__solc_builds is None:
            raise RuntimeError(
                f"Unable to fetch or correctly parse '{self.__solc_list_url}'."
            )

        if isinstance(version, str):
            version = SolidityVersion.fromstring(version)

        minimal_version = self.list_all()[0]
        if version < minimal_version:
            raise UnsupportedVersionError(
                f"The minimal supported solc version for the current platform is `{minimal_version}`."
            )

        if version not in self.__solc_builds.releases:
            raise ValueError(f"solc version `{version}` does not exist.")
        if self.get_path(version).is_file() and not force_reinstall:
            return

        filename = self.__solc_builds.releases[version]
        download_url = f"{self.BINARIES_URL}/{self.__platform}/{filename}"
        local_path = self.get_path(version).parent / filename

        local_path.parent.mkdir(parents=True, exist_ok=True)

        if http_session is None:
            async with aiohttp.ClientSession() as session:
                await self.__download_file(download_url, local_path, session, progress)
        else:
            await self.__download_file(download_url, local_path, http_session, progress)

        if not self.__verify_checksums(version):
            local_path.unlink()
            raise ChecksumError(
                f"Checksum of the downloaded solc version `{version}` does not match the expected value."
            )

        # unzip older Windows solc binary zipped together with DLLs
        if filename.endswith(".zip"):
            local_path = self.__unzip(local_path)

        local_path.chmod(0o775)

    def remove(self, version: Union[SolidityVersion, str]) -> None:
        path = self.get_path(version).parent
        if path.is_dir():
            shutil.rmtree(path)
        else:
            raise ValueError(
                f"solc version `{version}` was not installed - cannot remove."
            )

    def get_path(self, version: Union[SolidityVersion, str]) -> Path:
        self.__fetch_list_file()
        if self.__solc_builds is None:
            raise RuntimeError(
                f"Unable to fetch or correctly parse '{self.__solc_list_url}'."
            )

        if isinstance(version, str):
            version = SolidityVersion.fromstring(version)

        minimal_version = self.list_all()[0]
        if version < minimal_version:
            raise UnsupportedVersionError(
                f"The minimal supported solc version for the current platform is `{minimal_version}`."
            )

        if version not in self.__solc_builds.releases:
            raise ValueError(f"solc version `{version}` does not exist")

        filename = self.__solc_builds.releases[version]
        dirname = filename
        if dirname.endswith((".exe", ".zip")):
            dirname = dirname[:-4]
        if filename.endswith(".zip"):
            filename = filename[:-3] + "exe"
        return self.__compilers_path / dirname / filename

    def list_all(self) -> Tuple[SolidityVersion, ...]:
        self.__fetch_list_file()
        if self.__solc_builds is None:
            raise RuntimeError(
                f"Unable to fetch or correctly parse '{self.__solc_list_url}'."
            )

        return tuple(sorted(self.__solc_builds.releases.keys()))

    async def __download_file(
        self,
        url: str,
        path: Path,
        http_session: aiohttp.ClientSession,
        progress: Optional[Callable[[int, int], Awaitable[None]]] = None,
    ) -> None:
        async with http_session.get(url) as r:
            total_size = r.headers.get("Content-Length")
            if total_size is not None:
                total_size = int(total_size)
            downloaded_size = 0
            with path.open("wb") as f:
                async for chunk in r.content.iter_chunked(8 * 1024):
                    f.write(chunk)
                    downloaded_size += len(chunk)
                    if total_size is not None and progress is not None:
                        await progress(downloaded_size, total_size)

    def __unzip(self, zip_path: Path) -> Path:
        """
        Unzip the Windows `solc` executable zip containing:
        - solc.exe - extract this file and rename it as `solc-windows-amd64-v{version}+commit.{commit}.exe`
        - soltest.exe - ignore this file (i.e. do not extract it)
        - extract any additional files (DLLs) next to the solc binary
        After that, delete the zip file.
        """
        base_path = zip_path.parent
        solc_filename = zip_path.name[:-3] + "exe"
        solc_path = zip_path.parent / solc_filename

        with ZipFile(zip_path, "r") as _zip:
            members = _zip.namelist()
            for member in members:
                if member == "soltest.exe":
                    # do not extract soltest.exe to save up the space
                    continue
                elif member == "solc.exe":
                    # rename solc.exe to the long name (containing version number, commit number etc.)
                    _zip.extract(member, base_path)
                    (base_path / "solc.exe").rename(solc_path)
                else:
                    # extract all the remaining files
                    _zip.extract(member, base_path)
        zip_path.unlink()
        return solc_path

    def __fetch_list_file(self) -> None:
        """
        Download ``list.json`` file from `binaries.soliditylang.org <binaries.soliditylang.org>`_ for the current
        platform and save it as ``{global_data_path}/compilers/solc.json``. In case of network issues, try to
        use the locally downloaded solc builds file as a fallback.
        """
        if self.__solc_builds is not None:
            return

        try:
            with urllib.request.urlopen(self.__solc_list_url, timeout=0.25) as response:
                json = response.read()
                self.__solc_builds = SolcBuilds.parse_raw(json)
                self.__solc_list_path.write_bytes(json)
        except (urllib.error.URLError, OSError) as e:
            # in case of networking issues try to use the locally downloaded solc builds file as a fallback
            if self.__solc_list_path.is_file():
                self.__solc_builds = SolcBuilds.parse_file(self.__solc_list_path)
            else:
                raise

    def __verify_checksums(self, version: SolidityVersion) -> bool:
        assert self.__solc_builds is not None
        build_info = next(b for b in self.__solc_builds.builds if b.version == version)
        local_path = self.get_path(version)

        filename = self.__solc_builds.releases[version]
        if filename.endswith(".zip"):
            local_path = local_path.parent / filename
            if not local_path.is_file():
                return True

        sha256 = build_info.sha256
        if sha256.startswith("0x"):
            sha256 = sha256[2:]

        keccak256 = build_info.keccak256
        if keccak256.startswith("0x"):
            keccak256 = keccak256[2:]

        if not self.__verify_sha256(local_path, sha256):
            return False
        if not self.__verify_keccak256(local_path, keccak256):
            return False
        return True

    def __verify_sha256(self, path: Path, expected: str) -> bool:
        """
        Check SHA256 checksum of the provided file against the expected value.
        :param path: path of the file whose checksum to be verified
        :param expected: expected value of SHA256 checksum
        :return: True if checksum matches the expected value, False otherwise
        """
        h = hashlib.sha256()
        with path.open("rb") as f:
            while True:
                chunk = f.read(4 * 1024)
                if not chunk:
                    break
                h.update(chunk)

        return h.hexdigest() == expected

    def __verify_keccak256(self, path: Path, expected: str) -> bool:
        """
        Check KECCAK256 checksum of the provided file against the expected value.
        :param path: path of the file whose checksum to be verified
        :param expected: expected value of KECCAK256 checksum
        :return: True if checksum matches the expected value, False otherwise
        """
        h = keccak.new(digest_bits=256)
        with path.open("rb") as f:
            while True:
                chunk = f.read(4 * 1024)
                if not chunk:
                    break
                h.update(chunk)

        return h.hexdigest() == expected
